"""
Tagulous test: Model TagField

Modules tested:
    tagulous.models.managers.TagRelatedManager
    tagulous.models.descriptors.BaseTagDescriptor
    tagulous.models.descriptors.TagDescriptor
    tagulous.models.fields.BaseTagField
    tagulous.models.fields.TagField
"""
from django.db import models
from django.test import TestCase, TransactionTestCase

from tagulous import models as tag_models
from tests.lib import TagTestManager, skip_if_mysql
from tests.tagulous_tests_app import models as test_models


class ModelTagFieldTest(TagTestManager, TestCase):
    """
    Test model TagField
    """

    manage_models = [test_models.TagFieldModel]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldModel
        self.tag_model = test_models.TagFieldModel.tags.tag_model
        self.tag_field = test_models.TagFieldModel.tags

    def test_descriptor(self):
        "Check TagDescriptor is in place"
        self.assertIsInstance(self.tag_field, tag_models.TagDescriptor)

    def test_tag_table(self):
        "Check the tag table exists"
        self.assertTrue(
            issubclass(self.tag_field.field.remote_field.model, tag_models.TagModel)
        )
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))

    def test_empty_value(self):
        "Check the descriptor returns a TagRelatedManager"
        t1 = self.create(self.test_model, name="Test")
        self.assertInstanceEqual(t1, name="Test")

        # Check the TagDescriptor is returning a TagRelatedManager
        self.assertEqual(t1.tags.__class__.__name__, "TagRelatedManager")

        # Check it also has a reference to the correct model
        self.assertEqual(t1.tags.tag_model, self.tag_model)

    def test_tag_assign_before_save(self):
        """
        Check a tag string can be assigned to an instance which hasn't yet
        been saved
        """
        t1 = self.test_model(name="Test")
        t1.tags = "blue, red"
        self.assertEqual(t1.tags.get_tag_string(), "blue, red")
        self.assertTagModel(self.tag_model, {})

    def test_tag_assign_in_constructor(self):
        "Check a tag string can be set in the constructor"
        t1 = self.test_model(name="Test", tags="blue, red")

        # Returned before save
        self.assertEqual(t1.tags.get_tag_string(), "blue, red")
        self.assertTagModel(self.tag_model, {})

        # Returned after save
        t1.save()
        self.assertInstanceEqual(t1, name="Test", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

    def test_tag_assign_by_list_in_constructor(self):
        "Check a list of strings can be set in the constructor"
        t1 = self.test_model(name="Test", tags=["blue", "red"])

        # Returned before save
        self.assertEqual(t1.tags.get_tag_string(), "blue, red")
        self.assertTagModel(self.tag_model, {})

        # Returned after save
        t1.save()
        self.assertInstanceEqual(t1, name="Test", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_in_object_create(self):
        "Check a tag string can be set in object.create"
        t1 = self.test_model.objects.create(name="Test", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_by_list_in_object_create(self):
        "Check a list of strings can be set in object.create"
        t1 = self.test_model.objects.create(name="Test", tags=["blue", "red"])
        self.assertInstanceEqual(t1, name="Test", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_by_queryset_in_object_create(self):
        "Check a list of strings can be set in object.create"
        self.tag_model.objects.create(name="blue")
        self.tag_model.objects.create(name="red")
        t1 = self.test_model.objects.create(
            name="Test", tags=self.tag_model.objects.all()
        )
        self.assertInstanceEqual(t1, name="Test", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_in_object_get_or_create_true(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does not exist
        """
        t1, state = self.test_model.objects.get_or_create(name="Test", tags="blue, red")
        self.assertEqual(state, True)
        self.assertEqual(t1.name, "Test")
        self.assertEqual(t1.tags, "blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_in_object_get_or_create_false(self):
        """
        Check a tag string can be passed in object.get_or_create, when object
        does exist
        """
        t1 = self.test_model.objects.create(name="Test", tags="blue, red")
        t2 = self.test_model.objects.create(name="Test", tags="green")
        t3, state = self.test_model.objects.get_or_create(name="Test", tags="blue, red")
        self.assertEqual(state, False)
        self.assertEqual(t1.pk, t3.pk)
        self.assertInstanceEqual(t2, name="Test", tags="green")
        self.assertEqual(t3.name, "Test")
        self.assertEqual(t3.tags, "blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1, "green": 1})

    def test_tag_assign_string_obj_save(self):
        """
        Check a tag string can be assigned to the descriptor, saved when the
        instance is saved, and returned

        Checks TagRelatedManager post_save listener
        """
        t1 = self.create(self.test_model, name="Test 1")
        t1.tags = "red, blue"

        # Returned before save
        self.assertEqual(t1.tags.get_tag_string(), "blue, red")
        self.assertEqual(str(t1.tags), t1.tags.get_tag_string())
        self.assertEqual(len(t1.tags.get_tag_list()), 2)
        self.assertTrue("red" in t1.tags.get_tag_list())
        self.assertTrue("blue" in t1.tags.get_tag_list())
        self.assertTagModel(self.tag_model, {})

        # Check db hasn't changed
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.tags), "")

        # Returned after save
        t1.save()
        t1 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(t1.tags.get_tag_string(), "blue, red")
        self.assertEqual(str(t1.tags), t1.tags.get_tag_string())
        self.assertEqual(len(t1.tags.get_tag_list()), 2)
        self.assertTrue("red" in t1.tags.get_tag_list())
        self.assertTrue("blue" in t1.tags.get_tag_list())
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_string_manager_save(self):
        """
        Check a tag string can be assigned to the descriptor, saved by the tag
        manager, and returned

        Checks TagRelatedManager.save
        """
        t1 = self.create(self.test_model, name="Test 1")
        t1.tags = "red, blue"

        # Returned before save
        self.assertEqual(str(t1.tags), "blue, red")
        self.assertTagModel(self.tag_model, {})

        # Check db hasn't changed
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.tags), "")

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_assign_list_strings(self):
        "Check tags can be set with a list of strings"
        t1 = self.create(self.test_model, name="Test 1")
        t1.tags = ["red", "blue"]

        # Returned before save
        self.assertEqual(str(t1.tags), "blue, red")
        self.assertTagModel(self.tag_model, {})

        # Check db hasn't changed
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.tags), "")

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_assign_list_objs(self):
        "Check tags can be set with a list of strings"
        t1 = self.create(self.test_model, name="Test 1")
        t1.tags = [
            self.tag_model.objects.create(name="blue"),
            self.tag_model.objects.create(name="red"),
        ]

        # Returned before save
        self.assertEqual(str(t1.tags), "blue, red")
        self.assertTagModel(self.tag_model, {"red": 0, "blue": 0})

        # Check db hasn't changed
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.tags), "")

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_assign_queryset(self):
        "Check tags can be set with a queryset of tags"
        t1 = self.create(self.test_model, name="Test 1")
        self.tag_model.objects.create(name="blue")
        self.tag_model.objects.create(name="red")
        t1.tags = self.tag_model.objects.all()

        # Returned before save, check DB hasn't changed
        self.assertEqual(str(t1.tags), "blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertTagModel(self.tag_model, {"red": 0, "blue": 0})

        # Returned after save
        t1.tags.save()
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

    def test_tag_assign_string_empty(self):
        "Check setting an empty string clears tags"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})
        t1.tags = ""

        # Returned before save, check DB hasn't changed
        self.assertEqual(str(t1.tags), "")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertTagModel(self.tag_model, {})

    def test_assign_list_empty(self):
        "Check setting an empty list clears tags"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})
        t1.tags = []

        # Returned before save
        self.assertEqual(str(t1.tags), "")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

        # Check db hasn't changed
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.tags), "blue, red")

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertTagModel(self.tag_model, {})

    def test_assign_none(self):
        "Check setting None clears tags"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})
        t1.tags = None

        # Returned before save
        self.assertEqual(str(t1.tags), "")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

        # Check db hasn't changed
        t2 = self.test_model.objects.get(name="Test 1")
        self.assertEqual(str(t2.tags), "blue, red")

        # Returned after save
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertTagModel(self.tag_model, {})

    def test_assign_unknown(self):
        "Check assigning an unknown value raises an exception"
        t1 = self.create(self.test_model, name="Test 1")
        with self.assertRaises(ValueError) as cm:
            # Assign the descriptor to itself, because why not.
            t1.tags = self.test_model.tags
        self.assertEqual(str(cm.exception), "Unexpected value assigned to TagField")

    def test_contains(self):
        "Check __contains__ using strings and Tag objects"
        t1 = self.create(self.test_model, name="1", tags="blue, red")

        # String matches
        self.assertTrue("blue" in t1.tags, "blue not found")
        self.assertTrue("red" in t1.tags, "red not found")

        # Partials shouldn't match
        self.assertFalse("green" in t1.tags, "green incorrectly found")
        self.assertFalse("blu" in t1.tags, "blu incorrectly found")
        self.assertFalse("ed" in t1.tags, "ed incorrectly found")

        # Tag objects (will be cast to strings)
        red = t1.tags.tag_model.objects.get(name="red")
        blue = t1.tags.tag_model.objects.get(name="blue")
        green = t1.tags.tag_model.objects.create(name="green")
        self.assertTrue(red in t1.tags, "red not found")
        self.assertTrue(blue in t1.tags, "red not found")
        self.assertFalse(green in t1.tags, "green incorrectly found")

    def test_len(self):
        "Check manager count method returns the number of tags"
        t1 = self.create(self.test_model, name="1", tags="blue, red")
        self.assertEqual(t1.tags.count(), 2)

    def test_eq(self):
        "Check __eq__ correctly determines equality"
        t1 = self.create(self.test_model, name="1", tags="blue, red")
        self.assertEqual(t1.tags, "blue, red")
        self.assertEqual(t1.tags, "red, blue")
        self.assertEqual(t1.tags, ["red", "blue"])

        # Test queryset
        tags = t1.tags.tag_model.objects.all()
        self.assertEqual(t1.tags, tags)

        # Test another TagRelatedManager
        t2 = self.create(self.test_model, name="2", tags="blue, red")
        self.assertEqual(t1.tags, t2.tags)

    def test_ne(self):
        "Check __ne__ correctly determines falsity"
        t1 = self.create(self.test_model, name="1", tags="blue, red")
        self.assertNotEqual(t1.tags, "red")
        self.assertNotEqual(t1.tags, "blue")
        self.assertNotEqual(t1.tags, ["red", "green"])
        self.assertNotEqual(t1.tags, "green")

        # Test another TagRelatedManager
        t2 = self.create(self.test_model, name="2", tags="red, green")
        self.assertNotEqual(t1.tags, t2.tags)

        # Test queryset
        tags = t1.tags.tag_model.objects.all()
        self.assertNotEqual(t1.tags, tags)

    def test_multiple_instances(self):
        "Check multiple tagged instances work without interference"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        t2 = self.create(self.test_model, name="Test 2", tags="green, red")
        t3 = self.create(self.test_model, name="Test 3", tags="blue, green")

        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertInstanceEqual(t2, name="Test 2", tags="green, red")
        self.assertInstanceEqual(t3, name="Test 3", tags="blue, green")
        self.assertTagModel(self.tag_model, {"red": 2, "blue": 2, "green": 2})

    def test_change_string_add(self):
        "Add a tag by changing tag string"
        t1 = self.create(self.test_model, name="Test 1", tags="blue")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue")
        self.assertTagModel(self.tag_model, {"blue": 1})
        t1.tags = "blue, green"
        self.assertTagModel(self.tag_model, {"blue": 1})
        t1.tags.save()
        self.assertEqual(t1.tags, "blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

    def test_m2m_set_by_string(self):
        "Set a tag directly using M2M .set(string)"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})
        t1.tags.set("green", "pink")
        self.assertEqual(t1.tags, "green, pink")
        self.assertInstanceEqual(t1, name="Test 1", tags="green, pink")
        self.assertTagModel(self.tag_model, {"green": 1, "pink": 1})

        # No need to test further arguments, it passes through to add() which
        # is tested thoroughly below

    def test_m2m_add_by_obj(self):
        "Add a tag directly using M2M .add(obj)"
        t1 = self.create(self.test_model, name="Test 1", tags="blue")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue")
        self.assertTagModel(self.tag_model, {"blue": 1})
        t1.tags.add(self.tag_model.objects.create(name="green"))
        self.assertEqual(t1.tags, "blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

    def test_m2m_add_by_string(self):
        "Add a tag directly using M2M .add(str)"
        t1 = self.create(self.test_model, name="Test 1", tags="blue")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue")
        self.assertTagModel(self.tag_model, {"blue": 1})
        t1.tags.add("green")
        self.assertEqual(t1.tags, "blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

    def test_m2m_add_by_string_existing(self):
        "Add a tag directly using M2M .add(str) when the tag already exists"
        self.test_model.tags.tag_model.objects.create(name="green")
        t1 = self.create(self.test_model, name="Test 1", tags="blue")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 0})
        t1.tags.add("green")
        self.assertEqual(t1.tags, "blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

    def test_change_string_remove(self):
        "Remove a tag by changing tag string"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, green")
        t2 = self.create(self.test_model, name="Test 2", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 2, "green": 1, "red": 1})

        t1.tags = "green"
        self.assertTagModel(self.tag_model, {"blue": 2, "green": 1, "red": 1})
        t1.tags.save()
        self.assertInstanceEqual(t1, name="Test 1", tags="green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1, "red": 1})

    def test_m2m_remove_by_obj(self):
        "Remove a tag directly using M2M .remove(obj)"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, green")
        t2 = self.create(self.test_model, name="Test 2", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 2, "green": 1, "red": 1})

        t1.tags.remove(self.tag_model.objects.get(name="blue"))
        self.assertEqual(t1.tags, "green")
        self.assertInstanceEqual(t1, name="Test 1", tags="green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1, "red": 1})

    def test_m2m_remove_by_str(self):
        "Remove a tag directly using M2M .remove(str)"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, green")
        t2 = self.create(self.test_model, name="Test 2", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 2, "green": 1, "red": 1})

        t1.tags.remove("blue")
        self.assertEqual(t1.tags, "green")
        self.assertInstanceEqual(t1, name="Test 1", tags="green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1, "red": 1})

    def test_m2m_remove_not_set(self):
        "Remove a tag which exists but isn't set"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, green")
        t2 = self.create(self.test_model, name="Test 2", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 2, "green": 1, "red": 1})

        t1.tags.remove("blue", "green", "red")
        self.assertEqual(t1.tags, "")
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

    def test_m2m_remove_missing(self):
        "Remove a tag which doesn't exist (and isn't set)"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        t1.tags.remove("green")
        self.assertEqual(t1.tags, "blue, red")

    def test_m2m_clear(self):
        "Clear all tags from an item with manager.clear"
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        t2 = self.create(self.test_model, name="Test 2", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 2, "blue": 2})

        # Returned before save
        t1.tags.clear()
        self.assertEqual(str(t1.tags), "")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")

    def test_fake_manager_after_delete(self):
        "Check fake manager steps in to store tags when the item is deleted"
        # Create the item
        t1 = self.create(self.test_model, name="Test 1", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

        # Now delete it
        t1.delete()

        # It won't have a pk any more, but it will have a manager.
        self.assertEqual(t1.pk, None)

        # TagDescriptor will now have to create a FakeTagRelatedManager and
        # copy the tags across from the real one. Because although this is
        # about as likely as the previous test, someone will do it eventually.
        manager = t1.tags
        self.assertIsInstance(manager, tag_models.managers.FakeTagRelatedManager)
        self.assertEqual(manager, "blue, red")

    def test_fake_manager(self):
        "Check that the FakeTagRelatedManager doesn't do databases"
        errmsg = (
            '"<{model_name}: {model_name} object (None)>" needs to be '
            "saved before TagField can use the database"
        ).format(model_name=self.test_model._meta.object_name)

        t1 = self.test_model(name="Test 1", tags="blue, red")

        with self.assertRaises(ValueError) as cm:
            t1.tags.save()
        self.assertEqual(str(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.set("fail")
        self.assertEqual(str(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.add("fail")
        self.assertEqual(str(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.remove("blue")
        self.assertEqual(str(cm.exception), errmsg)

        with self.assertRaises(ValueError) as cm:
            t1.tags.clear()
        self.assertEqual(str(cm.exception), errmsg)

    def test_prefetch_related(self):
        self.create(self.test_model, name="Test 1", tags="blue, green")
        self.create(self.test_model, name="Test 2", tags="blue, red")
        self.create(self.test_model, name="Test 3", tags="blue, red")
        with self.assertNumQueries(2):
            [obj.tags for obj in self.test_model.objects.all().prefetch_related("tags")]


# ##############################################################################
# ######  Test it works with concrete inheritance
# ##############################################################################


class ModelTagFieldConcreteInheritanceTest(ModelTagFieldTest):
    manage_models = [test_models.TagFieldConcreteInheritanceModel]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldConcreteInheritanceModel
        self.tag_model = test_models.TagFieldConcreteInheritanceModel.tags.tag_model
        self.tag_field = test_models.TagFieldConcreteInheritanceModel.tags


# ##############################################################################
# ######  Test it works with abstract inheritance
# ##############################################################################


class ModelTagFieldAbstractInheritanceTest(ModelTagFieldConcreteInheritanceTest):
    manage_models = [test_models.TagFieldAbstractInheritanceModel]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldAbstractInheritanceModel
        self.tag_model = test_models.TagFieldAbstractInheritanceModel.tags.tag_model
        self.tag_field = test_models.TagFieldAbstractInheritanceModel.tags

    def test_second_inheritance__tag_model_not_shared(self):
        second_test_model = test_models.TagFieldAbstractInheritanceSecondModel
        second_tag_model = second_test_model.tags.tag_model

        self.assertNotEqual(self.tag_model, second_tag_model)

    def test_second_inheritance__writes_update_shared_model(self):
        second_test_model = test_models.TagFieldAbstractInheritanceSecondModel
        second_tag_model = second_test_model.tags.tag_model

        t1 = self.test_model.objects.create(name="Test 1", tags="blue, red")
        t2 = second_test_model.objects.create(name="Test 2", tags="green, yellow")

        self.assertInstanceEqual(t1, name="Test 1", tags="blue, red")
        self.assertTagModel(self.tag_model, {"red": 1, "blue": 1})

        self.assertInstanceEqual(t2, name="Test 2", tags="green, yellow")
        self.assertTagModel(second_tag_model, {"green": 1, "yellow": 1})


# ##############################################################################
# ######  Test invalid fields
# ##############################################################################


class ModelTagFieldInvalidTest(TagTestManager, TransactionTestCase):
    """
    Test invalid model TagField

    Use a TransactionTestCase so the apps will be reset
    """

    manage_models = [test_models.SingleTagFieldModel]

    def test_to_model_options_fails(self):
        "Check when a to model is specified, tag options are invalid"
        new_field = tag_models.TagField(test_models.TagMetaUser, force_lowercase=True)
        with self.assertRaises(ValueError) as cm:
            new_field._process_deferred_options()
        self.assertEqual(
            str(cm.exception),
            "Cannot set tag options on explicit tag model "
            "<class 'tests.tagulous_tests_app.models.TagMetaUser'>",
        )

    def test_to_model_force_options_works(self):
        "Check when _set_tag_meta is True, options can be set after all"
        self.assertTrue(test_models.TagMetaModel.tag_options.force_lowercase)
        new_field = tag_models.TagField(
            test_models.TagMetaModel, force_lowercase=False, _set_tag_meta=True
        )
        new_field._process_deferred_options()
        self.assertFalse(test_models.TagMetaModel.tag_options.force_lowercase)

        # Set force_lowercase back ready for future tests - unittest won't fix
        test_models.TagMetaModel.tag_options.force_lowercase = True
        self.assertTrue(test_models.TagMetaModel.tag_options.force_lowercase)

    def test_invalid_to_model(self):
        "Check that the to model has to be a TagModel subclass"
        with self.assertRaises(ValueError) as cm:

            class FailModel_invalid_to(models.Model):
                to_model = tag_models.TagField(test_models.TagFieldModel)

        self.assertEqual(str(cm.exception), "Tag model must be a subclass of TagModel")

    def test_forbidden_db_table(self):
        "Check that db_table argument raises exception"
        with self.assertRaises(ValueError) as cm:

            class FailModel_forbidden_db(models.Model):
                db_table = tag_models.TagField(db_table="fail")

        self.assertEqual(str(cm.exception), "Invalid argument 'db_table' for TagField")

    def test_forbidden_through(self):
        "Check that through argument raises exception"
        with self.assertRaises(ValueError) as cm:

            class FailModel_forbidden_through(models.Model):
                through = tag_models.TagField(through="fail")

        self.assertEqual(str(cm.exception), "Invalid argument 'through' for TagField")

    def test_forbidden_symmetrical(self):
        "Check that symmetrical argument raises exception"
        with self.assertRaises(ValueError) as cm:

            class FailModel_forbidden_symmetrical(models.Model):
                symmetrical = tag_models.TagField(symmetrical="fail")

        self.assertEqual(
            str(cm.exception), "Invalid argument 'symmetrical' for TagField"
        )

    def test_nulled_tag_field(self):
        "Check model field raises warning when given invalid null"
        nulled_tagfield = tag_models.TagField(null=True)
        warnings = nulled_tagfield._check_ignored_options()
        self.assertEqual(len(warnings), 1)
        self.assertEqual(warnings[0].msg, "null has no effect on TagField.")


# ##############################################################################
# ###### Test tag counter
# ##############################################################################


class ModelTagFieldCountTest(TagTestManager, TestCase):
    """
    Test tag model counter
    """

    manage_models = [test_models.TagFieldModel]

    def setUpExtra(self):
        self.tag_model = test_models.TagFieldModel.tags.tag_model
        self.tag_field = test_models.TagFieldModel.tags

    def test_count_0_create(self):
        "Check an unprotected tag can be created with a tag count of 0"
        self.tag_model.objects.create(name="red")
        self.assertTagModel(self.tag_model, {"red": 0})

    def test_count_0_delete(self):
        "Check that an unprotected tag is deleted when count changes to 0"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

        t1.tags.remove("blue")
        self.assertInstanceEqual(t1, name="Test 1", tags="green")
        self.assertTagModel(self.tag_model, {"green": 1})

    def test_instance_delete_updates(self):
        "Check that deleting an instance updates the count"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags="blue, green")
        t2 = self.create(test_models.TagFieldModel, name="Test 2", tags="blue, red")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 2, "green": 1, "red": 1})

        t1.delete()
        self.assertInstanceEqual(t2, name="Test 2", tags="blue, red")
        self.assertTagModel(self.tag_model, {"blue": 1, "red": 1})

    def test_protected_remains(self):
        "Check that a protected tag stays when count changes to 0"
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

        green = self.tag_model.objects.get(name="green")
        green.protected = True
        green.save()

        t1.delete()
        self.assertTagModel(self.tag_model, {"green": 0})

    def test_clear_count_0_delete(self):
        """
        Check that an unprotected tag is deleted when clear changes count to 0,
        but protected tag remains
        """
        t1 = self.create(test_models.TagFieldModel, name="Test 1", tags="blue, green")
        self.assertInstanceEqual(t1, name="Test 1", tags="blue, green")
        self.assertTagModel(self.tag_model, {"blue": 1, "green": 1})

        green = self.tag_model.objects.get(name="green")
        green.protected = True
        green.save()

        t1.tags.clear()
        self.assertInstanceEqual(t1, name="Test 1", tags="")
        self.assertTagModel(self.tag_model, {"green": 0})


# ##############################################################################
# ######  Test model field blank=True
# ##############################################################################


class ModelTagFieldOptionalTest(TagTestManager, TestCase):
    """
    Test optional TagField
    """

    manage_models = [test_models.TagFieldOptionalModel]

    def test_optional_save_missing(self):
        "Check an optional TagField isn't required for save"
        # If it fails, it will return an exception
        t1 = test_models.TagFieldOptionalModel(name="Test 1")
        t1.save()
        self.assertNotEqual(t1.pk, None)

    def test_optional_create_missing(self):
        "Check an optional TagField isn't required for object.create"
        # If it fails, it will return an exception
        t1 = test_models.TagFieldOptionalModel.objects.create(name="Test 1")
        self.assertNotEqual(t1.pk, None)


# ##############################################################################
# ###### Test model field blank=False
# ##############################################################################


class ModelTagFieldRequiredTest(TagTestManager, TestCase):
    """
    Test required model TagField
    """

    manage_models = [test_models.TagFieldRequiredModel]

    def test_required_save(self):
        "Check a required TagField doesn't raise an exception when saved"
        # This isn't exactly the desired action, but as it stands we can't
        # enforce blank=False on a M2M tagfield without rewriting chunks of
        # django's core modelform - it can't even be monkeypatched cleanly.
        #
        # Future changes to django may make it possible; at that point we can
        # switch this test to enforce blank=False:
        """
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.TagFieldRequiredModel(name='Test')
            t1.save()
        self.assertEqual(cm.exception.messages[0], 'This field cannot be null.')
        """
        # but until then:
        t1 = test_models.TagFieldRequiredModel(name="Test")
        t1.save()
        self.assertInstanceEqual(t1, name="Test", tag="")

    def test_required_create_raises(self):
        "Check a required TagField raises an exception in object.create"
        # As test_required_save, this isn't desired behaviour, but until django
        # changes, we can't use this test:
        """
        with self.assertRaises(exceptions.ValidationError) as cm:
            t1 = test_models.TagFieldRequiredModel.objects.create(name='Test')
            t1.full_clean()
        self.assertEqual(cm.exception.messages[0], 'This field cannot be null.')
        """
        # instead:
        t1 = test_models.TagFieldRequiredModel.objects.create(name="Test")
        self.assertInstanceEqual(t1, name="Test", tag="")


# ##############################################################################
# ###### Test multiple TagFields on a model
# ##############################################################################


class ModelTagFieldMultipleTest(TagTestManager, TestCase):
    """
    Test multiple tag fields on a model
    """

    manage_models = [test_models.TagFieldMultipleModel]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldMultipleModel
        self.tag_field_1 = test_models.TagFieldMultipleModel.tags1
        self.tag_field_2 = test_models.TagFieldMultipleModel.tags2
        self.tag_field_3 = test_models.TagFieldMultipleModel.tags3

    def test_tag_models(self):
        self.assertTrue(issubclass(self.tag_field_1.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_2.tag_model, tag_models.TagModel))
        self.assertTrue(issubclass(self.tag_field_3.tag_model, tag_models.TagModel))

    def test_separate_models(self):
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_2.tag_model)
        self.assertNotEqual(self.tag_field_1.tag_model, self.tag_field_3.tag_model)
        self.assertNotEqual(self.tag_field_2.tag_model, self.tag_field_3.tag_model)

    def test_model_names(self):
        self.assertEqual(
            self.tag_field_1.tag_model.__name__, "Tagulous_TagFieldMultipleModel_tags1"
        )
        self.assertEqual(
            self.tag_field_2.tag_model.__name__, "Tagulous_TagFieldMultipleModel_tags2"
        )
        self.assertEqual(
            self.tag_field_3.tag_model.__name__, "Tagulous_TagFieldMultipleModel_tags3"
        )

    def test_set_and_get(self):
        "Test multiple fields can be set and retrieved independently"
        t1 = self.create(
            self.test_model,
            name="Test 1",
            tags1="adam, brian",
            tags2="alice, brenda",
            tags3="aragorn, bilbo",
        )
        t2 = self.create(
            self.test_model,
            name="Test 2",
            tags1="chris, david",
            tags2="claire, daphne",
            tags3="celeborn, denethor",
        )
        t3 = self.create(
            self.test_model,
            name="Test 3",
            tags1="eric, frank",
            tags2="edith, faye",
            tags3="eowyn, frodo",
        )

        self.assertTagModel(
            self.tag_field_1,
            {"adam": 1, "brian": 1, "chris": 1, "david": 1, "eric": 1, "frank": 1},
        )
        self.assertTagModel(
            self.tag_field_2,
            {"alice": 1, "brenda": 1, "claire": 1, "daphne": 1, "edith": 1, "faye": 1},
        )
        self.assertTagModel(
            self.tag_field_3,
            {
                "aragorn": 1,
                "bilbo": 1,
                "celeborn": 1,
                "denethor": 1,
                "eowyn": 1,
                "frodo": 1,
            },
        )

        self.assertInstanceEqual(
            t1,
            name="Test 1",
            tags1="adam, brian",
            tags2="alice, brenda",
            tags3="aragorn, bilbo",
        )
        self.assertInstanceEqual(
            t2,
            name="Test 2",
            tags1="chris, david",
            tags2="claire, daphne",
            tags3="celeborn, denethor",
        )
        self.assertInstanceEqual(
            t3,
            name="Test 3",
            tags1="eric, frank",
            tags2="edith, faye",
            tags3="eowyn, frodo",
        )


# ##############################################################################
# ###### Test TagField with string references to tag model
# ##############################################################################


class ModelTagFieldStringTest(TagTestManager, TransactionTestCase):
    """
    Test tag field which refers to its tag model with a string
    """

    manage_models = [test_models.MixedStringTo]

    def setUpExtra(self):
        self.test_model = test_models.MixedStringTo
        self.tag_field = self.test_model.tags
        self.tag_model = test_models.MixedStringTagModel

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        self.assertEqual(self.tag_field.field.remote_field.model, self.tag_model)
        self.assertEqual(self.tag_field.tag_model, self.tag_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(self.tag_field.tag_options, self.tag_model.tag_options)

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.tag_model, {})
        t1 = self.test_model.objects.create(name="Test 1", tags="one, two")
        self.assertTagModel(self.tag_model, {"one": 1, "two": 1})
        self.assertInstanceEqual(t1, name="Test 1", tags="one, two")


class ModelTagFieldSelfTest(TagTestManager, TransactionTestCase):
    """
    Test tag field which refers to itself
    """

    manage_models = [test_models.MixedSelfTo]

    def setUpExtra(self):
        self.test_model = test_models.MixedSelfTo
        self.tag_field = self.test_model.related

    def test_to_model(self):
        "Check related model is correct"
        self.assertTrue(issubclass(self.tag_field.tag_model, tag_models.TagModel))
        self.assertEqual(self.tag_field.field.remote_field.model, self.test_model)
        self.assertEqual(self.tag_field.tag_model, self.test_model)

    def test_tag_options(self):
        "Check tag options are available correctly"
        self.assertEqual(self.tag_field.tag_options, self.test_model.tag_options)

    def test_use(self):
        "Test basic use of tag field"
        self.assertTagModel(self.test_model, {})
        t1 = self.test_model.objects.create(name="Test 1", related="one, two")
        self.assertTagModel(self.test_model, {"one": 1, "two": 1, "Test 1": 0})
        self.assertInstanceEqual(t1, name="Test 1", related="one, two")


# ##############################################################################
# ###### Test TagField options
# ##############################################################################


class ModelTagFieldOptionsTest(TagTestManager, TransactionTestCase):
    """
    Test tag field options
    """

    manage_models = [test_models.TagFieldOptionsModel]

    def setUpExtra(self):
        self.test_model = test_models.TagFieldOptionsModel

    def test_initial_string(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(
            self.test_model.initial_string, {"Adam": 0, "Brian": 0, "Chris": 0}
        )

    def test_initial_list(self):
        # Initial will have been loaded by TagTestManager
        self.assertTagModel(
            self.test_model.initial_list, {"Adam": 0, "Brian": 0, "Chris": 0}
        )

    def test_protect_initial_true(self):
        self.assertTagModel(self.test_model.protect_initial_true, {"Adam": 0})

        t1 = self.create(self.test_model, name="Test 1", protect_initial_true="Adam")
        self.assertTagModel(self.test_model.protect_initial_true, {"Adam": 1})

        t1.protect_initial_true = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_true, {"Adam": 0})

    def test_protect_initial_false(self):
        self.assertTagModel(self.test_model.protect_initial_true, {"Adam": 0})

        t1 = self.create(self.test_model, name="Test 1", protect_initial_false="Adam")
        self.assertTagModel(self.test_model.protect_initial_false, {"Adam": 1})

        t1.protect_initial_false = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_initial_false, {})

    def test_protect_all_true(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_true="Adam")
        self.assertTagModel(self.test_model.protect_all_true, {"Adam": 1})

        t1.protect_all_true = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_all_true, {"Adam": 0})

    def test_protect_all_false(self):
        t1 = self.create(self.test_model, name="Test 1", protect_all_false="Adam")
        self.assertTagModel(self.test_model.protect_all_false, {"Adam": 1})

        t1.protect_all_false = ""
        t1.save()
        self.assertTagModel(self.test_model.protect_all_false, {})

    @skip_if_mysql
    def test_case_sensitive_true(self):
        self.assertTagModel(self.test_model.case_sensitive_true, {"Adam": 0})
        self.create(self.test_model, name="Test 1", case_sensitive_true="adam")
        self.assertTagModel(self.test_model.case_sensitive_true, {"Adam": 0, "adam": 1})

    @skip_if_mysql
    def test_case_sensitive_false(self):
        self.assertTagModel(self.test_model.case_sensitive_false, {"Adam": 0})
        self.create(self.test_model, name="Test 1", case_sensitive_false="adam")
        self.assertTagModel(self.test_model.case_sensitive_false, {"Adam": 1})

    @skip_if_mysql
    def test_cmp_case_sensitive_true(self):
        "Test case sensitive matches"
        t1 = self.create(
            self.test_model, name="Test 1", case_sensitive_true="django, html"
        )
        self.assertEqual(t1.case_sensitive_true, "django, html")
        self.assertNotEqual(t1.case_sensitive_true, "django, HTML")
        self.assertNotEqual(t1.case_sensitive_true, "Django, html")

    @skip_if_mysql
    def test_contains_case_sensitive_true(self):
        "Check case sensitive __contains__"
        t1 = self.create(
            self.test_model, name="Test 1", case_sensitive_true="Django, HTML"
        )

        # String matches
        self.assertTrue("Django" in t1.case_sensitive_true)
        self.assertTrue("HTML" in t1.case_sensitive_true)

        # Partials and incorrect shouldn't be found
        self.assertFalse("Javascript" in t1.case_sensitive_true)
        self.assertFalse("django" in t1.case_sensitive_true)
        self.assertFalse("Html" in t1.case_sensitive_true)

    @skip_if_mysql
    def test_cmp_case_sensitive_false(self):
        """
        Test case insensitive matches

        Note: force_lowercase is also false here
        """
        t1 = self.create(
            self.test_model, name="Test 1", case_sensitive_false="Adam, Brian"
        )
        self.assertEqual(t1.case_sensitive_false, "Adam, Brian")
        self.assertEqual(t1.case_sensitive_false, "adam, BRIAN")
        self.assertNotEqual(t1.case_sensitive_false, "Chris")

    def test_force_lowercase_true(self):
        self.create(self.test_model, name="Test 1", force_lowercase_true="Adam")
        self.assertTagModel(self.test_model.force_lowercase_true, {"adam": 1})

    def test_force_lowercase_false(self):
        self.create(self.test_model, name="Test 1", force_lowercase_false="Adam")
        self.assertTagModel(self.test_model.force_lowercase_false, {"Adam": 1})

    def test_cmp_force_lowercase_true(self):
        "Test matches when force lowercase true and case sensitive is false"
        t1 = self.create(
            self.test_model, name="Test 1", force_lowercase_true="blue, red"
        )
        self.assertEqual(t1.force_lowercase_true, "blue, red")
        self.assertEqual(t1.force_lowercase_true, "Blue, RED")
        self.assertNotEqual(t1.force_lowercase_true, "green")

    def test_contains_force_lowercase_true(self):
        "Check __contains__ when force lowercase true and case sensitive false"
        t1 = self.create(
            self.test_model, name="Test 1", force_lowercase_true="django, html"
        )

        # String matches are forced to lowercase
        self.assertTrue("Django" in t1.force_lowercase_true)
        self.assertTrue("HTML" in t1.force_lowercase_true)
        self.assertTrue("django" in t1.force_lowercase_true)
        self.assertTrue("html" in t1.force_lowercase_true)

        # Incorrect shouldn't be found
        self.assertFalse("Javascript" in t1.case_sensitive_true)

    def test_cmp_case_sensitive_true_force_lowercase_true(self):
        "Test matches when case sensitive and force lowercase true"
        t1 = self.create(self.test_model, name="Test 1", cs_true_fl_true="blue, red")
        self.assertEqual(t1.cs_true_fl_true, "blue, red")
        self.assertEqual(t1.cs_true_fl_true, "Blue, RED")
        self.assertNotEqual(t1.cs_true_fl_true, "green")

    def test_max_count_below(self):
        t1 = self.create(self.test_model, name="Test 1", max_count="Adam")
        self.assertInstanceEqual(t1, name="Test 1", max_count="Adam")
        self.assertTagModel(self.test_model.max_count, {"Adam": 1})

    def test_max_count_equal(self):
        t1 = self.create(self.test_model, name="Test 1", max_count="Adam, Brian, Chris")
        self.assertInstanceEqual(t1, name="Test 1", max_count="Adam, Brian, Chris")
        self.assertTagModel(
            self.test_model.max_count, {"Adam": 1, "Brian": 1, "Chris": 1}
        )

    def test_max_count_create_above(self):
        with self.assertRaises(ValueError) as cm:
            self.test_model.objects.create(
                name="Test 1", max_count="Adam, Brian, Chris, David"
            )
        self.assertEqual(str(cm.exception), "Cannot set more than 3 tags on this field")
        with self.assertRaises(self.test_model.DoesNotExist) as cm:
            self.test_model.objects.get(name="Test 1")
        self.assertTagModel(self.test_model.max_count, {})

    def test_max_count_assign_equal_replaces(self):
        t1 = self.create(self.test_model, name="Test 1", max_count="Adam, Brian, Chris")
        t1.max_count = "Adam, Brian, David"
        t1.save()
        self.assertInstanceEqual(t1, name="Test 1", max_count="Adam, Brian, David")
        self.assertTagModel(
            self.test_model.max_count, {"Adam": 1, "Brian": 1, "David": 1}
        )

    def test_max_count_assign_above(self):
        t1 = self.create(self.test_model, name="Test 1")
        with self.assertRaises(ValueError) as cm:
            t1.max_count = "Adam, Brian, Chris, David"
        self.assertEqual(str(cm.exception), "Cannot set more than 3 tags on this field")
        self.assertInstanceEqual(t1, name="Test 1", max_count="")
        self.assertTagModel(self.test_model.max_count, {})

    def test_max_count_add_above(self):
        t1 = self.create(self.test_model, name="Test 1", max_count="Adam, Brian")
        with self.assertRaises(ValueError) as cm:
            t1.max_count.add("Chris", "David")
        self.assertEqual(
            str(cm.exception),
            "Cannot set more than 3 tags on this field; it already has 2",
        )
        self.assertInstanceEqual(t1, name="Test 1", max_count="Adam, Brian")
        # They won't have been created
        self.assertTagModel(self.test_model.max_count, {"Adam": 1, "Brian": 1})



# -*- coding: utf-8 -*-
from django.db import migrations, models

import tagulous.models.fields
import tagulous.models.models


class Migration(migrations.Migration):

    dependencies = [("tagulous_tests_migration", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Tagulous_MigrationTestModel_singletag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(unique=True, max_length=191)),
                ("slug", models.SlugField()),
                (
                    "count",
                    models.IntegerField(
                        default=0,
                        help_text=b"Internal counter of how many times this tag is in use",
                    ),
                ),
                (
                    "protected",
                    models.BooleanField(
                        default=False,
                        help_text=b"Will not be deleted when the count reaches 0",
                    ),
                ),
            ],
            options={"ordering": ("name",), "abstract": False},
            bases=(tagulous.models.models.BaseTagModel, models.Model),
        ),
        migrations.CreateModel(
            name="Tagulous_MigrationTestModel_tags",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(unique=True, max_length=191)),
                ("slug", models.SlugField()),
                (
                    "count",
                    models.IntegerField(
                        default=0,
                        help_text=b"Internal counter of how many times this tag is in use",
                    ),
                ),
                (
                    "protected",
                    models.BooleanField(
                        default=False,
                        help_text=b"Will not be deleted when the count reaches 0",
                    ),
                ),
            ],
            options={"ordering": ("name",), "abstract": False},
            bases=(tagulous.models.models.BaseTagModel, models.Model),
        ),
        migrations.AlterUniqueTogether(
            name="tagulous_migrationtestmodel_tags", unique_together=set([("slug",)])
        ),
        migrations.AlterUniqueTogether(
            name="tagulous_migrationtestmodel_singletag",
            unique_together=set([("slug",)]),
        ),
        migrations.AddField(
            model_name="migrationtestmodel",
            name="singletag",
            field=tagulous.models.fields.SingleTagField(
                force_lowercase=False,
                to="tagulous_tests_migration.Tagulous_MigrationTestModel_singletag",
                blank=True,
                help_text=b"Enter a comma-separated tag string",
                null=True,
                autocomplete_initial=True,
                autocomplete_view=b"tagulous_tests_app-null",
                autocomplete_limit=3,
                initial="Mr",
                protect_all=False,
                case_sensitive=True,
                protect_initial=True,
                _set_tag_meta=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="migrationtestmodel",
            name="tags",
            field=tagulous.models.fields.TagField(
                force_lowercase=True,
                to="tagulous_tests_migration.Tagulous_MigrationTestModel_tags",
                protect_initial=True,
                help_text=b"Enter a comma-separated tag string",
                case_sensitive=True,
                autocomplete_initial=True,
                autocomplete_view=b"tagulous_tests_app-null",
                autocomplete_limit=3,
                initial="Mr",
                protect_all=False,
                _set_tag_meta=True,
            ),
            preserve_default=True,
        ),
    ]



"""
Tagulous test: Model field order

Modules tested:
    tagulous.models.fields.SingleTagField
    tagulous.models.fields.TagField
"""
from django.test import TestCase

from tests.lib import TagTestManager
from tests.tagulous_tests_app import models as test_models


class ModelFieldOrderTest(TagTestManager, TestCase):
    """
    Test model SingleTagField order
    """

    manage_models = [test_models.MixedOrderTest]

    def test_order_correct(self):
        """
        Check that the order of the non-ManyToMany fields is correct
        This is to check that Django internals haven't changed significantly
        """
        # Check the ordering is as expected
        opts = test_models.MixedOrderTest._meta
        local_fields = sorted(opts.concrete_fields + opts.many_to_many)
        expected_fields = [
            # Auto pk 'id'
            "id",
            # Defined fields
            "char1",
            "fk1",
            "char2",
            "single1",
            "char3",
            "m2m1",
            "char4",
            "multi1",
            "char5",
            "m2m2",
            "char6",
            "fk2",
            "char7",
        ]
        self.assertEqual(len(local_fields), len(expected_fields))
        for i in range(len(local_fields)):
            self.assertEqual(local_fields[i].name, expected_fields[i])



"""
Tag parsing and printing

Loosely based on django-taggit and django-tagging
"""
import unicodedata

from django.utils.encoding import force_str

from .constants import COMMA, DOUBLE_QUOTE, QUOTE, SPACE, TREE


try:
    from unidecode import unidecode
except ImportError:  # pragma: no cover - tests simulate this
    unidecode = None


# ##############################################################################
# ###### Tag name parse and render
# ##############################################################################


def parse_tags(tag_string, max_count=0, space_delimiter=True):
    """
    Tag parser

    Rules without quotes:
        If a comma is present it's used as the delimiter
        Otherwise space is used as the delimiter
        Spaces at the start and end of tags are ignored

    Rules with quotes
        Quotes can be escaped by double quotes, ie ""
        Commas outside quotes take precedence over spaces as delimiter
        Unmatched quotes will be left in the string

    If space_delimiter is False, space will never be used as a delimiter.

    Tree tags can be further split into their parts with split_tree_name
    """
    # Empty string easiest case
    if not tag_string:
        return []

    tag_string = force_str(tag_string)

    # Prep variables for the parser
    tags = []
    tag = ""
    delimiter = SPACE
    in_quote = None
    chars = False

    # Disable spaces
    if not space_delimiter:
        delimiter = COMMA

    # Bypass main parser for efficiency if no quotes
    if QUOTE not in tag_string:
        # No quotes - simple split and strip

        # Normally split on commas
        delimiter = COMMA

        # But if no commas, split on spaces
        if COMMA not in tag_string and space_delimiter:
            delimiter = SPACE

        # Split and strip tags
        tags = split_strip(tag_string, delimiter)

    else:
        # Break tag string into list of (index, char)
        chars = list(enumerate(tag_string))

    # Loop through chars
    while chars:
        index, char = chars.pop(0)

        # See if it's a delimiter
        if not in_quote:
            # Comma delimiter takes priority
            if delimiter != COMMA and char == COMMA:
                delimiter = COMMA

                # All previous tags were actually just one tag
                tag = tag_string[0:index].strip()
                tags = []

                # Strip start/end quotes
                tag_len = len(tag)
                tag = tag.lstrip(QUOTE)
                left_quote_count = tag_len - len(tag)
                tag_len = len(tag)
                tag = tag.rstrip(QUOTE)
                right_quote_count = tag_len - len(tag)

                # Escape inner quotes
                tag = tag.replace(DOUBLE_QUOTE, QUOTE)

                # Add back escaped start/end quotes
                tag = (
                    (QUOTE * int(left_quote_count / 2))
                    + tag
                    + (QUOTE * int(right_quote_count / 2))
                )

                # Add back insignificant unescaped quotes.
                #
                # There are only two scenarios where there can be unescaped
                # quotes at the start, followed by a comma later:
                #   1. The comma is quoted - but that means in_quote is True,
                #      in which case we won't be in this code branch
                #   2. The comma comes after a matching closing unescaped quote
                #
                # Therefore there can't be insigificant unescaped quotes on the
                # left and unescaped quotes on the right are only insignificant
                # if there are no unescaped quotes on the left
                if right_quote_count % 2 == 1 and left_quote_count % 2 == 0:
                    tag += QUOTE

            # Found end of tag
            if char == delimiter:
                tag = tag.rstrip()
                if tag:
                    tags.append(tag)
                    tag = ""
                # Following tested manually due to coverage bug
                #   See https://bitbucket.org/ned/coveragepy/issues/198
                continue  # pragma: no cover

            # If tag is empty, ignore whitespace
            if not tag and char == SPACE:
                continue

        # Now either in a quote, or not a delimiter
        # If it's not a quote, add to tag
        if char != QUOTE:
            tag += char
            continue

        # Char is quote - count how many quotes appear here
        quote_count = 1
        while chars and chars[0][1] == QUOTE:
            quote_count += 1
            chars.pop(0)

        if not tag:
            # Quote at start
            # If an odd number, now in quote
            if quote_count % 2 == 1:
                in_quote = True

            # Tag starts with escaped quotes
            tag = QUOTE * int(quote_count / 2)
        else:
            # Quote in middle or at end
            # Add any escaped
            tag += QUOTE * int(quote_count / 2)

            # An odd number followed by a delimiter will mean it has ended
            # Need to look ahead to figure it out
            if quote_count % 2 == 1:

                # If it's the last character, it has closed
                if len(chars) == 0:
                    in_quote = False
                    break

                for i2, c2 in chars:
                    if c2 == SPACE:
                        if delimiter == SPACE:
                            # Quotes closed; tag will end next loop
                            in_quote = False
                            break
                        else:
                            # Spaces are insignificant during whitespace
                            # Tag may continue, keep checking chars
                            # Following tested manually due to coverage bug
                            continue  # pragma: no cover
                    elif c2 == COMMA:
                        # Quotes closed; tag will end next loop
                        # Delimiter doesn't matter, comma always wins
                        in_quote = False
                        break

                    # Tag has not ended
                    # Add odd quote to tag and keep building
                    tag += QUOTE
                    break

    # Chars expended
    if tag:
        # Partial tag remains; add to stack
        if in_quote:
            # Add the quote back to the start - it wasn't significant after all
            tag = QUOTE + tag
        tags.append(tag)

    # Enforce uniqueness and sort
    tags = list(set(tags))
    tags.sort()

    # Check the count
    if max_count and len(tags) > max_count:
        raise ValueError(
            "This field can only have %s argument%s"
            % (max_count, "" if max_count == 1 else "s")
        )

    return tags


def split_strip(string, delimiter=","):
    """
    Splits ``string`` on ``delimiter``, stripping each resulting string
    and returning a list of non-empty strings.
    """
    if not string:
        return []

    words = [w.strip() for w in string.split(delimiter)]
    return [w for w in words if w]


def render_tags(tags):
    """
    Creates a tag string from a list of Tag instances or strings, suitable for
    editing.

    Tag names which contain commas will be quoted, existing quotes will be
    escaped.
    """
    names = []
    for tag in tags:
        # This will catch a list of Tag objects or tag name strings
        name = str(tag)

        name = name.replace(QUOTE, DOUBLE_QUOTE)
        if COMMA in name or SPACE in name:
            names.append('"%s"' % name)
        else:
            names.append(name)
    return ", ".join(sorted(names))


# ##############################################################################
# ###### Tree name split and join
# ##############################################################################


def split_tree_name(name):
    """
    Split a tree tag name into its parts

    A slash can be escaped by double slash, ie //
    """
    name = name.strip()
    if len(name) == 0:
        return []

    # Replace escaped slashes with escape char, split on remaining slashes
    ESCAPE_CHAR = "\0"
    escaped = name.replace(TREE + TREE, ESCAPE_CHAR)
    return [x.strip().replace(ESCAPE_CHAR, TREE) for x in escaped.split(TREE)]


def join_tree_name(parts):
    """
    Join tree tag name parts into a single name string

    A slash in a part will be escaped by double slash, ie //
    """
    return TREE.join(part.replace(TREE, TREE + TREE) for part in parts)


def clean_tree_name(name):
    """
    Make sure a tree name is valid

    * Escapes leading or trailing slashes
    * Strips leading and trailing whitespace around parts
    """
    # Count the number of slashes, ensure it's even (all escaped)
    left = len(name) - len(name.lstrip(TREE))
    if left % 2 == 1:
        name = TREE + name

    right = len(name) - len(name.rstrip(TREE))
    if right % 2 == 1:
        name += TREE

    # Split and join to strip whitespace
    return join_tree_name(split_tree_name(name))



from django import forms
from django.contrib import admin, messages
from django.core.exceptions import ImproperlyConfigured
from django.db.models.base import ModelBase
from django.http import HttpResponseRedirect
from django.shortcuts import render

from . import forms as tag_forms
from . import models as tag_models


# ##############################################################################
# ########################################################## Admin classes
# ##############################################################################


class TaggedBaseModelAdminMixin(admin.options.BaseModelAdmin):
    """
    Mixin for BaseModelAdmin subclasses which are for tagged models
    """

    def formfield_for_dbfield(self, db_field, **kwargs):
        """
        Remove the RelatedFieldWidgetWrapper from tag fields, so they don't
        display popup buttons
        """
        formfield = super(TaggedBaseModelAdminMixin, self).formfield_for_dbfield(
            db_field, request=kwargs.pop("request", None), **kwargs
        )

        if (
            isinstance(db_field, (tag_models.SingleTagField, tag_models.TagField))
            and isinstance(formfield.widget, admin.widgets.RelatedFieldWidgetWrapper)
            and isinstance(formfield.widget.widget, tag_forms.AdminTagWidget)
        ):
            formfield.widget = formfield.widget.widget
        return formfield

    def get_autocomplete_fields(self, request):
        """
        Ensure TagFields aren't listed in Django's autocomplete_fields - they don't
        play well with Django's autocomplete, and will use their own
        """
        autocomplete_fields = super().get_autocomplete_fields(request)
        safe_fields = [
            field
            for field in autocomplete_fields
            if not isinstance(
                self.opts.get_field(field),
                (tag_models.SingleTagField, tag_models.TagField),
            )
        ]
        return safe_fields


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   ModelAdmin for Tagged models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TaggedModelAdmin(TaggedBaseModelAdminMixin, admin.ModelAdmin):
    pass


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   ModelAdmin for TagModel models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagModelAdmin(admin.ModelAdmin):
    list_display = ["name", "count", "protected"]
    list_filter = ["protected"]
    search_fields = ["name"]
    exclude = ["count"]
    actions = ["merge_tags"]
    prepopulated_fields = {"slug": ("name",)}

    def merge_tags(self, request, queryset):
        """
        Admin action to merge tags
        """
        # Thanks to:
        #   http://www.hoboes.com/Mimsy/hacks/django-actions-their-own-intermediate-page/

        # Create a form
        is_tree = issubclass(self.model, tag_models.TagTreeModel)

        class MergeForm(forms.Form):
            # Keep selected items in same field, admin.helpers.ACTION_CHECKBOX_NAME
            _selected_action = forms.CharField(widget=forms.MultipleHiddenInput)
            # Allow use to select from selected items
            merge_to = forms.ModelChoiceField(queryset)

        if is_tree:

            class MergeForm(MergeForm):
                # Allow to merge recursively
                merge_children = forms.BooleanField(required=False)

        if "merge" in request.POST:
            merge_form = MergeForm(request.POST)
            if merge_form.is_valid():
                # Merge - with children if set
                merge_to = merge_form.cleaned_data["merge_to"]
                kwargs = {}
                if is_tree and merge_form.cleaned_data["merge_children"]:
                    kwargs["children"] = True
                merge_to.merge_tags(queryset, **kwargs)

                self.message_user(request, "Tags merged", messages.SUCCESS)
                return HttpResponseRedirect(request.get_full_path())

        else:
            tag_pks = request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME)
            if len(tag_pks) < 2:
                self.message_user(
                    request,
                    "You must select at least two tags to merge",
                    messages.ERROR,
                )
                return HttpResponseRedirect(request.get_full_path())

            merge_form = MergeForm(
                initial={
                    admin.helpers.ACTION_CHECKBOX_NAME: request.POST.getlist(
                        admin.helpers.ACTION_CHECKBOX_NAME
                    ),
                    "merge_children": True,
                }
            )

        return render(
            request,
            "tagulous/admin/merge_tags.html",
            {
                "title": "Merge tags",
                "opts": self.model._meta,
                "merge_form": merge_form,
                "tags": queryset,
            },
        )

    merge_tags.short_description = "Merge selected tags..."


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   ModelAdmin for TagTreeModel models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagTreeModelAdmin(TagModelAdmin):
    exclude = ["count", "parent", "path", "label", "level"]


# ##############################################################################
# ########################################################## Admin registration
# ##############################################################################

# Give contrib.admin a default widget for tag fields
admin.options.FORMFIELD_FOR_DBFIELD_DEFAULTS.update(
    {
        tag_models.SingleTagField: {"widget": tag_forms.AdminTagWidget},
        tag_models.TagField: {"widget": tag_forms.AdminTagWidget},
    }
)


def _create_display(field):
    """
    ModelAdmin display function factory
    """

    def display(self, obj):
        return getattr(obj, field).get_tag_string()

    display.short_description = field.replace("_", " ")
    return display


def enhance(model, admin_class):
    """
    Add tag support to the admin class based on the specified model
    """
    #
    # Get a list of all tag fields
    #

    # Dict of single tag fields, {name: tag}
    single_tag_fields = {}

    # Dict of normal tag fields, {name: tag}
    tag_fields = {}

    # List of all single and normal tag fields
    tag_field_names = []

    # Check for SingleTagField related fields
    for field in tag_models.singletagfields_from_model(model):
        single_tag_fields[field.name] = field
        tag_field_names.append(field.name)

    # Check for TagField m2m fields
    for field in tag_models.tagfields_from_model(model):
        tag_fields[field.name] = field
        tag_field_names.append(field.name)

    #
    # Ensure any tag fields in list_display are rendered by functions
    #

    # The admin.site.register will complain if it's a ManyToManyField, so this
    # will work around that.
    #
    # We also need to have a different name to the model field, otherwise the
    # ChangeList class will just use the model field - that would get the tag
    # strings showing in the table, but the column would be sortable which
    # would cause problems for TagFields, and the display function would never
    # get called, which would be unexpected for anyone maintaining this code.
    if hasattr(admin_class, "list_display"):
        # Make sure we're working with a list
        if isinstance(admin_class.list_display, tuple):
            admin_class.list_display = list(admin_class.list_display)

        for i, field in enumerate(admin_class.list_display):
            # If the field's not a callable, and not in the admin class already
            if not hasattr(field, "__call__") and not hasattr(admin_class, field):
                # Only TagFields (admin can already handle SingleTagField FKs)
                if field in tag_fields:
                    # Create new field name and replace in list_display
                    display_name = "_tagulous_display_%s" % field
                    admin_class.list_display[i] = display_name

                    # Add display function to admin class
                    setattr(admin_class, display_name, _create_display(field))

    #
    # If admin is for a tag model, ensure any inlines for tagged models are
    # subclasses of TaggedInlineFormSet.
    #
    if issubclass(model, tag_models.BaseTagModel) and hasattr(admin_class, "inlines"):
        for inline_cls in admin_class.inlines:
            # Make sure inline class uses TaggedBaseModelAdminMixin
            if not issubclass(inline_cls, TaggedBaseModelAdminMixin):
                inline_cls.__bases__ = (
                    TaggedBaseModelAdminMixin,
                ) + inline_cls.__bases__

            # Make sure inlines used TaggedInlineFormSet
            if issubclass(inline_cls.model, tag_models.TaggedModel) and not issubclass(
                inline_cls.formset, tag_forms.TaggedInlineFormSet
            ):
                orig_cls = inline_cls.formset
                inline_cls.formset = type(
                    str("Tagged%s" % orig_cls.__name__),
                    (tag_forms.TaggedInlineFormSet, orig_cls),
                    {},
                )


def register(model, admin_class=None, site=None, **options):
    """
    Provide tag support to the model when it is registered with the admin site.

    For tagged models (have one or more SingleTagField or TagField fields):
        * Admin will support TagField in list_display

    For tag models (subclass of TagModel):
        * Admin will provide a merge action to merge tags

    For other models:
        * No changes made

    Arguments:
        model       Model to register
        admin_class Admin class for model
        site        Admin site to register with
                    Default: django.contrib.admin.site
        **options   Extra options for admin class

    This only supports one model, but is otherwise safe to use with non-tagged
    models.
    """
    # Look at the model we've been given
    if isinstance(model, tag_models.BaseTagDescriptor):
        # It's a tag descriptor; change it for the tag model itself
        model = model.tag_model

    elif not isinstance(model, ModelBase):
        raise ImproperlyConfigured(
            "Tagulous can only register a single model with admin."
        )

    # Ensure we have a valid admin site
    if site is None:
        site = admin.site

    #
    # Determine appropriate admin class
    #
    if not admin_class:
        admin_class = admin.ModelAdmin

    # Going to make a list of base classes to inject
    cls_bases = []

    # If it's a tag model, subclass TagModelAdmin or TagTreeModelAdmin
    if issubclass(model, tag_models.BaseTagModel):
        if issubclass(model, tag_models.TagTreeModel):
            if not issubclass(admin_class, TagTreeModelAdmin):
                cls_bases += [TagTreeModelAdmin]
        else:
            if not issubclass(admin_class, TagModelAdmin):
                cls_bases += [TagModelAdmin]

    # If it's a tagged model, subclass TaggedModelAdmin
    singletagfields = tag_models.singletagfields_from_model(model)
    tagfields = tag_models.tagfields_from_model(model)
    if singletagfields or tagfields:
        if not issubclass(admin_class, TaggedModelAdmin):
            cls_bases += [TaggedModelAdmin]

    # If options specified, or other bases, will need to subclass admin_class
    if options or cls_bases:
        cls_bases += [admin_class]
        # Update options with anything the new subclasses could have overidden
        # in a custom ModelAdmin - unless they're already overridden in options
        options["__module__"] = __name__
        if admin_class != admin.ModelAdmin:
            options.update(
                dict(
                    (k, v)
                    for k, v in admin_class.__dict__.items()
                    if k in ["list_display", "list_filter", "exclude", "actions"]
                    and k not in options
                )
            )
        admin_class = type(str("%sAdmin" % model.__name__), tuple(cls_bases), options)

    # Enhance the model admin class
    enhance(model, admin_class)

    # Register the model
    # Don't pass options - we've already dealt with that
    site.register(model, admin_class)



"""
Tagulous test: Admin

Modules tested:
    tagulous.admin
"""
import copy
import re

import django
from django.contrib import admin, messages
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.messages.storage.fallback import CookieStorage
from django.core import exceptions
from django.http import HttpRequest, QueryDict
from django.test import TestCase
from django.urls import get_resolver, re_path, reverse

from tagulous import admin as tag_admin
from tagulous import forms as tag_forms
from tests.lib import TagTestManager
from tests.tagulous_tests_app import admin as test_admin
from tests.tagulous_tests_app import models as test_models
from tests.tagulous_tests_app import urls as test_urls


MOCK_PATH = "mock/path"


class TestRequestMixin(object):
    def mock_request(self, GET=None, POST=None):
        """
        Create a fake Request object based on the GET and POST kwargs
        """
        r = HttpRequest()
        r.path = MOCK_PATH
        r.method = "POST" if POST is not None else "GET"
        r.GET = GET or QueryDict("")
        r.POST = POST or QueryDict("")
        r._messages = CookieStorage(r)
        r.user = User.objects.get_or_create(username="test_user")[0]
        return r


class AdminTestManager(object):
    """
    Add per-test site urls to global urlpatterns
    """

    def setUp(self):
        super(AdminTestManager, self).setUp()

        if not hasattr(self, "site"):
            return

        # Try to clear the resolver cache
        if hasattr(get_resolver, "cache_clear"):
            # Django 2.2
            get_resolver.cache_clear()
        else:
            # Django 3.0+
            from django.urls.resolvers import _get_cached_resolver

            _get_cached_resolver.cache_clear()

        # Store the old urls and make a copy
        self.old_urls = test_urls.urlpatterns
        test_urls.urlpatterns = copy.copy(test_urls.urlpatterns)

        # Add the site to the copy
        test_urls.urlpatterns += [
            re_path(r"^tagulous_tests_app/admin/", self.site.urls)
        ]

    def tearDown(self):
        super(AdminTestManager, self).tearDown()
        if not hasattr(self, "old_urls"):
            return

        # Restore the original urls
        test_urls.urlpatterns = self.old_urls


# ##############################################################################
# ###### Admin registration
# ##############################################################################


class AdminRegisterTest(TestRequestMixin, TagTestManager, TestCase):
    """
    Test Admin registration of tagged model
    """

    def setUpExtra(self):
        self.admin = test_admin.SimpleMixedTestAdmin
        self.model = test_models.SimpleMixedTest
        self.model_singletag = self.model.singletag.tag_model
        self.model_tags = self.model.tags.tag_model
        self.site = admin.AdminSite(name="tagulous_admin")

    def test_register_site(self):
        "Check register with site"
        self.assertFalse(self.model in self.site._registry)
        tag_admin.register(self.model, self.admin, site=self.site)
        self.assertTrue(self.model in self.site._registry)
        ma = self.site._registry[self.model]
        self.assertIsInstance(ma, tag_admin.TaggedModelAdmin)

    def test_register_no_site(self):
        "Check register without site"
        # Replace admin.site with our own
        old_admin_site = admin.site
        admin.site = self.site
        self.assertFalse(self.model in self.site._registry)
        tag_admin.register(self.model, self.admin)
        self.assertTrue(self.model in self.site._registry)
        ma = self.site._registry[self.model]
        self.assertIsInstance(ma, tag_admin.TaggedModelAdmin)

        # Return admin.site
        admin.site = old_admin_site

    def test_register_models(self):
        "Check register refuses multiple models"
        with self.assertRaises(exceptions.ImproperlyConfigured) as cm:
            tag_admin.register([self.model, self.model], self.admin, site=self.site)
        self.assertEqual(
            str(cm.exception), "Tagulous can only register a single model with admin.",
        )
        self.assertFalse(self.model in self.site._registry)

    def test_register_auto(self):
        "Check register without admin dynamically creates admin class"
        self.assertFalse(self.model in self.site._registry)
        tag_admin.register(self.model, site=self.site)
        self.assertTrue(self.model in self.site._registry)
        ma = self.site._registry[self.model]
        self.assertIsInstance(ma, tag_admin.TaggedModelAdmin)

    def test_register_options(self):
        "Check register with options dynamically creates admin class"
        self.assertFalse(self.model in self.site._registry)
        tag_admin.register(
            self.model, self.admin, site=self.site, list_display=["name"]
        )
        self.assertTrue(self.model in self.site._registry)
        ma = self.site._registry[self.model]
        self.assertIsInstance(ma, tag_admin.TaggedModelAdmin)
        self.assertSequenceEqual(ma.get_list_display(self.mock_request()), ["name"])

    def test_register_tag_descriptor(self):
        "Check register tag descriptor creates correct admin class"
        self.assertFalse(self.model_singletag in self.site._registry)
        tag_admin.register(self.model.singletag, self.admin, site=self.site)
        self.assertTrue(self.model_singletag in self.site._registry)
        ma = self.site._registry[self.model_singletag]
        self.assertIsInstance(ma, tag_admin.TagModelAdmin)

    def test_register_tag_model(self):
        "Check register tag model creates correct admin class"
        self.assertFalse(self.model_singletag in self.site._registry)
        tag_admin.register(self.model_singletag, site=self.site)
        self.assertTrue(self.model_singletag in self.site._registry)
        ma = self.site._registry[self.model_singletag]
        self.assertIsInstance(ma, tag_admin.TagModelAdmin)

    def test_register_tag_model_class_properties(self):
        """
        Check list_display, list_filter, exclude and actions can be set in a
        tag modeladmin which doesn't explicitly subclass TagModelAdmin
        """
        # register class SimpleMixedTestTagsAdmin(admin.ModelAdmin):
        # against self.model
        self.assertFalse(self.model.singletag.tag_model in self.site._registry)
        tag_admin.register(
            self.model.singletag.tag_model,
            test_admin.SimpleMixedTestTagsAdmin,
            site=self.site,
        )
        self.assertTrue(self.model.singletag.tag_model in self.site._registry)
        ma = self.site._registry[self.model.singletag.tag_model]
        self.assertIsInstance(ma, tag_admin.TagModelAdmin)
        self.assertEqual(ma.list_display, ["name"])
        self.assertEqual(ma.list_filter, ["count"])
        self.assertEqual(ma.exclude, ["name"])
        self.assertEqual(ma.actions, [])

    def test_register_tag_tree_model(self):
        "Check register tag tree model creates correct admin class"
        tag_model = test_models.TreeTest.tags.tag_model
        self.assertFalse(tag_model in self.site._registry)
        tag_admin.register(tag_model, site=self.site)
        self.assertTrue(tag_model in self.site._registry)
        ma = self.site._registry[tag_model]
        self.assertIsInstance(ma, tag_admin.TagModelAdmin)


# ##############################################################################
# ###### Tagged ModelAdmin
# ##############################################################################


class TaggedAdminTest(TestRequestMixin, AdminTestManager, TagTestManager, TestCase):
    """
    Test ModelAdmin enhancements
    """

    def setUpExtra(self):
        self.admin = test_admin.SimpleMixedTestAdmin
        self.model = test_models.SimpleMixedTest
        self.model_singletag = self.model.singletag.tag_model
        self.model_tags = self.model.tags.tag_model
        self.site = admin.AdminSite(name="tagulous_admin")
        tag_admin.register(self.model, self.admin, site=self.site)
        self.ma = self.site._registry[self.model]
        self.cl = None

    def get_changelist(self, req=None):
        if not req:
            req = self.mock_request()
        list_display = self.ma.get_list_display(req)
        list_display_links = self.ma.get_list_display_links(req, list_display)
        list_filter = self.ma.get_list_filter(req)
        ChangeList = self.ma.get_changelist(req)
        changelist_args = [
            req,
            self.model,
            list_display,
            list_display_links,
            list_filter,
            self.ma.date_hierarchy,
            self.ma.search_fields,
            self.ma.list_select_related,
            self.ma.list_per_page,
            self.ma.list_max_show_all,
            self.ma.list_editable,
            self.ma,
            self.ma.get_sortable_by(req),
        ]
        self.cl = ChangeList(*changelist_args)
        return self.cl

    def get_changelist_results(self, req=None):
        if not req:
            req = self.mock_request()
        self.get_changelist(req)
        self.cl.get_results(req)
        return self.cl.result_list

    #
    # Tests
    #

    def test_changelist_display(self):
        "Check display fields have registered ok and return valid values"
        t1 = self.model.objects.create(name="Test 1", singletag="Mr", tags="red, blue")
        self.assertSequenceEqual(
            self.ma.get_list_display(self.mock_request()),
            ["name", "singletag", "_tagulous_display_tags"],
        )
        results = self.get_changelist_results()
        self.assertEqual(len(results), 1)
        r1 = results[0]
        self.assertEqual(t1.pk, r1.pk)

        # Find what it's showing
        from django.contrib.admin.templatetags.admin_list import items_for_result

        row = [list(items_for_result(self.cl, result, None)) for result in results][0]

        # Before comparing, strip class attrs
        row = [re.sub(r' class=".+?"', "", r) for r in row]
        self.assertSequenceEqual(
            row, ["<td>Test 1</td>", "<td>Mr</td>", "<td>blue, red</td>"]
        )

    def test_changelist_filter(self):
        t1 = self.model.objects.create(name="Test 1", singletag="Mr", tags="red, blue")
        t2 = self.model.objects.create(
            name="Test 2", singletag="Mrs", tags="red, green"
        )
        t3 = self.model.objects.create(
            name="Test 3", singletag="Mr", tags="green, blue"
        )

        # Check filters are listed
        self.assertSequenceEqual(
            self.ma.get_list_filter(self.mock_request()), ["singletag", "tags"]
        )

        # Filter by singletag
        singletag_request = self.mock_request(
            GET={"singletag__id__exact": self.model_singletag.objects.get(name="Mr").pk}
        )
        results = sorted(
            self.get_changelist_results(singletag_request), key=lambda r: r.name
        )
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].pk, t1.pk)
        self.assertEqual(results[1].pk, t3.pk)

        # Filter by tag
        tag_request = self.mock_request(
            GET={"tags__id__exact": self.model_tags.objects.get(name="green").pk}
        )
        results = sorted(self.get_changelist_results(tag_request), key=lambda r: r.name)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0].pk, t2.pk)
        self.assertEqual(results[1].pk, t3.pk)

    def test_form_fields(self):
        "Check forms get the correct widget"
        db_singletag = self.model._meta.get_field("singletag")
        field_singletag = self.ma.formfield_for_dbfield(db_singletag)
        self.assertIsInstance(field_singletag.widget, tag_forms.AdminTagWidget)

        db_tags = self.model._meta.get_field("tags")
        field_tags = self.ma.formfield_for_dbfield(db_tags)
        self.assertIsInstance(field_tags.widget, tag_forms.AdminTagWidget)


# ##############################################################################
# ###### Tag model admin tools
# ##############################################################################


class TagAdminTestManager(TestRequestMixin, AdminTestManager, TagTestManager, TestCase):
    """
    Test Admin registration of a tag model
    """

    def setUpModels(self):
        raise NotImplementedError()

    def setUpExtra(self):
        self.setUpModels()
        self.site = admin.AdminSite(name="tagulous_admin")
        tag_admin.register(self.model, site=self.site)
        self.ma = self.site._registry[self.model]
        self.cl = None

    def get_changelist(self, req=None):
        if not req:
            req = self.mock_request()
        list_display = self.ma.get_list_display(req)
        list_display_links = self.ma.get_list_display_links(req, list_display)
        list_filter = self.ma.get_list_filter(req)
        ChangeList = self.ma.get_changelist(req)
        changelist_args = [
            req,
            self.model,
            list_display,
            list_display_links,
            list_filter,
            self.ma.date_hierarchy,
            self.ma.search_fields,
            self.ma.list_select_related,
            self.ma.list_per_page,
            self.ma.list_max_show_all,
            self.ma.list_editable,
            self.ma,
            self.ma.get_sortable_by(req),
        ]
        self.cl = ChangeList(*changelist_args)
        return self.cl

    def assertContains(self, content, *seeks):
        for seek in seeks:
            self.assertTrue(seek in content, msg="Missing %s" % seek)

    def assertNotContains(self, content, *seeks):
        for seek in seeks:
            self.assertFalse(seek in content, msg="Unexpected %s" % seek)

    def do_test_merge_form(self, tags, excluded_tags, is_tree=False):
        "Request the form view and check it returns valid expected HTML"
        request = self.mock_request(
            POST=QueryDict(
                "&".join(
                    ["action=merge_tags"]
                    + [
                        "%s=%s" % (admin.helpers.ACTION_CHECKBOX_NAME, tag.pk)
                        for tag in tags
                    ]
                )
            )
        )
        cl = self.get_changelist(request)
        response = self.ma.response_action(request, cl.get_queryset(request))
        msgs = list(messages.get_messages(request))

        # Check response is appropriate
        self.assertEqual(len(msgs), 0)
        content = response.content.decode("utf-8")
        content_form = content[content.index("<form ") : content.index("</form>") + 7]

        # Strip csrfmiddlewaretoken
        content_form = re.sub(
            r"""<input [^>]*name=["']csrfmiddlewaretoken["'][^>]*>""", "", content_form
        )

        # Check HTML up to first <option> tag
        self.assertHTMLEqual(
            content_form[: content_form.index("<option")],
            (
                '<form action="" method="POST">'
                '<p><label for="id_merge_to">Merge to:</label>'
                '<select id="id_merge_to" {{required}}name="merge_to">'
            ),
        )

        # Can't be sure of options order
        options_raw = content_form[
            content_form.index("<option") : content_form.index("</select>")
        ]
        options = [
            "<option %s" % opt.strip() for opt in options_raw.split("<option ") if opt
        ]
        self.assertTrue('<option value="" selected>---------</option>' in options)

        for tag in tags:
            self.assertContains(
                options, '<option value="%d">%s</option>' % (tag.pk, tag.name)
            )

        # Find remaining input tags
        inputs_raw = content_form[
            # First input, convenient
            content_form.index("<input") : content_form.index("<div>")
        ]

        # If it's a tree, it should have merge_children first
        if is_tree:
            merge_children_label = (
                '<label for="id_merge_children">Merge children:</label>'
            )
            self.assertContains(content_form, merge_children_label)
            merge_children_input, inputs_raw = inputs_raw.split(">", 1)
            self.assertHTMLEqual(
                merge_children_input + ">",
                (
                    '<input type="checkbox" name="merge_children" '
                    'id="id_merge_children" checked="checked">'
                ),
            )

        # Remaining input tags should be tag ids selected on previous page
        inputs = [
            "<input %s" % inpt.strip()
            for inpt in
            # Some versions of django may insert a </p> here.
            inputs_raw.replace("</p>", "").strip().split("<input ")
            if inpt
        ]
        for action_id in range(len(tags)):
            self.assertTrue(
                any(
                    'id="id__selected_action_%s"' % action_id in input_tag
                    for input_tag in inputs
                )
            )

        # Now check the tag IDs are all there
        found_tag_pks = []
        for input_tag in inputs:
            # First check tag is expected format (numbers change)
            self.assertHTMLEqual(
                re.sub(r"\d+", "X", input_tag),
                (
                    '<input id="id__selected_action_X" type="hidden"'
                    ' name="_selected_action" value="X" />'
                ),
            )
            # Now find ID
            matches = re.search(r'value="(\d+)"', input_tag)
            if not matches:
                raise ValueError("Could not find tag pk in %s" % input_tag)
            found_tag_pks.append(int(matches.group(1)))
        self.assertEqual(set(found_tag_pks), set([tag.pk for tag in tags]))

        # Check end of form
        self.assertHTMLEqual(
            content_form[content_form.index("<div>") : content_form.index("</div>")],
            (
                "<div>"
                '<input type="submit" name="merge" value="Merge tags">'
                '<input class="default" type="hidden" name="action"'
                ' value="merge_tags">'
                "</div>"
            ),
        )

        # And just confirm excluded tags weren't there
        for tag in excluded_tags:
            self.assertNotContains(
                content, '<option value="%d">%s</option>' % (tag.pk, tag.name)
            )

    def do_form_submit(self, tags, merge_to, params=None):
        "Submit the form and check it succeeds and returns valid expected HTML"
        if params is None:
            params = []
        request = self.mock_request(
            POST=QueryDict(
                "&".join(
                    [
                        # Submitting
                        "action=merge_tags",
                        "merge=Merge%%20Tags",
                    ]
                    + [
                        # These were selected on the changelist, the ones we're merging
                        "%s=%s" % (admin.helpers.ACTION_CHECKBOX_NAME, tag.pk)
                        for tag in tags
                    ]
                    + [
                        # This is the one we're merging to
                        "merge_to=%s"
                        % merge_to.pk
                    ]
                    + params
                )
            )
        )
        cl = self.get_changelist(request)
        response = self.ma.response_action(request, cl.get_queryset(request))
        msgs = list(messages.get_messages(request))

        # Check response is appropriate
        self.assertEqual(len(msgs), 1)
        self.assertEqual(msgs[0].message, "Tags merged")

        if django.VERSION >= (3, 2):
            location = response.headers["location"]
        else:
            location = response._headers["location"][1]
        self.assertEqual(location, MOCK_PATH)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   TagModel admin tools
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagAdminTest(TagAdminTestManager, TestRequestMixin):
    """
    Test TagModel admin tools
    """

    def setUpModels(self):
        self.tagged_model = test_models.SimpleMixedTest
        self.model = self.tagged_model.tags.tag_model

    def populate(self):
        self.o1 = self.tagged_model.objects.create(
            name="Test 1", singletag="Mr", tags="red, blue"
        )
        self.o2 = self.tagged_model.objects.create(
            name="Test 2", singletag="Mrs", tags="red, green"
        )
        self.o3 = self.tagged_model.objects.create(
            name="Test 3", singletag="Mr", tags="green, blue"
        )
        self.red = self.model.objects.get(name="red")
        self.blue = self.model.objects.get(name="blue")
        self.green = self.model.objects.get(name="green")

        self.assertTagModel(self.model, {"red": 2, "blue": 2, "green": 2})

    def test_merge_action(self):
        "Check merge_tags action exists"
        self.assertTrue("merge_tags" in self.ma.actions)
        request = self.mock_request()
        actions = self.ma.get_actions(request)
        self.assertTrue("merge_tags" in actions)
        self.assertTrue(hasattr(actions["merge_tags"][0], "__call__"))
        self.assertEqual(actions["merge_tags"][1], "merge_tags")
        self.assertEqual(actions["merge_tags"][2], "Merge selected tags...")

    def test_merge_form_empty(self):
        "Check the merge_tags action fails when no tags selected"
        request = self.mock_request(POST=QueryDict("action=merge_tags"))
        cl = self.get_changelist(request)
        response = self.ma.response_action(request, cl.get_queryset(request))
        msgs = list(messages.get_messages(request))

        self.assertEqual(len(msgs), 1)
        self.assertEqual(
            msgs[0].message,
            (
                "Items must be selected in order to perform actions on them. No "
                "items have been changed."
            ),
        )
        self.assertEqual(response, None)

    def test_merge_form_one(self):
        "Check the merge_tags action fails when only one tag selected"
        self.populate()
        request = self.mock_request(
            POST=QueryDict(
                "action=merge_tags&%s=%s"
                % (admin.helpers.ACTION_CHECKBOX_NAME, self.red.pk)
            )
        )
        cl = self.get_changelist(request)
        response = self.ma.response_action(request, cl.get_queryset(request))
        msgs = list(messages.get_messages(request))

        self.assertEqual(len(msgs), 1)
        self.assertEqual(msgs[0].message, "You must select at least two tags to merge")

        if django.VERSION >= (3, 2):
            location = response.headers["location"]
        else:
            location = response._headers["location"][1]
        self.assertEqual(location, MOCK_PATH)

    def test_merge_form_two(self):
        "Check the merge_tags form when two tag selected"
        self.populate()
        self.do_test_merge_form(
            tags=[self.red, self.green], excluded_tags=[self.blue], is_tree=False
        )

    def test_merge_form_submit(self):
        "Check the merge_tag form submits and merges"
        self.populate()
        self.do_form_submit(tags=[self.red, self.green], merge_to=self.red)

        # Check they're merged
        self.assertTagModel(self.model, {"red": 3, "blue": 2})
        self.assertInstanceEqual(
            self.o1, name="Test 1", singletag="Mr", tags="blue, red"
        )
        self.assertInstanceEqual(self.o2, name="Test 2", singletag="Mrs", tags="red")
        self.assertInstanceEqual(
            self.o3, name="Test 3", singletag="Mr", tags="blue, red"
        )


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#   TagTreeModel admin tools
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagTreeAdminTest(TagAdminTestManager):
    """
    Test admin tools specific to TagTreeModel
    """

    def setUpModels(self):
        self.tagged_model = test_models.CustomTreeTest
        self.model = test_models.CustomTagTree

    def populate(self):
        # Normal Animal leaf nodes
        self.tagged_model.objects.create(name="cat1", tags="Animal/Mammal/Cat")
        self.tagged_model.objects.create(name="dog1", tags="Animal/Mammal/Dog")
        self.tagged_model.objects.create(name="bee1", tags="Animal/Insect/Bee")

        # Pet tree has Cat leaf and extra L2 not in Animal
        self.tagged_model.objects.create(name="cat2", tags="Pet/Mammal/Cat")
        self.tagged_model.objects.create(name="robin1", tags="Pet/Bird/Robin")

        # Food tree has Cat leaf and extra L3 not in Animal
        self.tagged_model.objects.create(name="cat3", tags="Food/Mammal/Cat")
        self.tagged_model.objects.create(name="pig1", tags="Food/Mammal/Pig")

        self.assertTagModel(
            self.model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 1,
                "Animal/Mammal/Dog": 1,
                "Pet": 0,
                "Pet/Mammal": 0,
                "Pet/Mammal/Cat": 1,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
                "Food": 0,
                "Food/Mammal": 0,
                "Food/Mammal/Cat": 1,
                "Food/Mammal/Pig": 1,
            },
        )

    def test_merge_form(self):
        """
        Check merge_tags adds merge_children option
        """
        self.populate()
        tag_names = ["Animal/Mammal", "Pet/Mammal", "Food/Mammal"]
        excluded_tags = self.model.objects.exclude(name__in=tag_names)
        self.assertEqual(
            excluded_tags.count(), self.model.objects.all().count() - len(tag_names)
        )

        self.do_test_merge_form(
            tags=self.model.objects.filter(name__in=tag_names),
            excluded_tags=excluded_tags,
            is_tree=True,
        )

    def test_merge_form_l2_without_children_submit(self):
        self.populate()

        # Tag something with 'Pet/Mammal'
        t1 = self.tagged_model.objects.create(name="mammal1", tags="Pet/Mammal")
        self.assertEqual(self.model.objects.get(name="Pet/Mammal").count, 1)

        # Now merge
        tag_names = ["Animal/Mammal", "Pet/Mammal", "Food/Mammal"]
        self.do_form_submit(
            tags=self.model.objects.filter(name__in=tag_names),
            merge_to=self.model.objects.get(name="Animal/Mammal"),
        )

        self.assertTagModel(
            self.model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 1,
                "Animal/Mammal/Cat": 1,
                "Animal/Mammal/Dog": 1,
                "Pet": 0,
                "Pet/Mammal": 0,
                "Pet/Mammal/Cat": 1,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
                "Food": 0,
                "Food/Mammal": 0,
                "Food/Mammal/Cat": 1,
                "Food/Mammal/Pig": 1,
            },
        )
        self.assertInstanceEqual(t1, name="mammal1", tags="Animal/Mammal")

    def test_merge_form_l2_with_children_submit(self):
        self.populate()
        tag_names = ["Animal/Mammal", "Pet/Mammal", "Food/Mammal"]
        self.do_form_submit(
            tags=self.model.objects.filter(name__in=tag_names),
            merge_to=self.model.objects.get(name="Animal/Mammal"),
            params=["merge_children=on"],
        )

        self.assertTagModel(
            self.model,
            {
                "Animal": 0,
                "Animal/Insect": 0,
                "Animal/Insect/Bee": 1,
                "Animal/Mammal": 0,
                "Animal/Mammal/Cat": 3,
                "Animal/Mammal/Dog": 1,
                "Animal/Mammal/Pig": 1,
                "Pet": 0,
                "Pet/Bird": 0,
                "Pet/Bird/Robin": 1,
            },
        )


# ##############################################################################
# ###### Tagged inlines on tag ModelAdmin
# ##############################################################################


class TaggedInlineSingleAdminTest(AdminTestManager, TagTestManager, TestCase):
    """
    Test inlines of tagged models on tag modeladmins
    """

    admin_cls = test_admin.SimpleMixedTestSingletagAdmin
    tagged_model = test_models.SimpleMixedTest
    model = test_models.SimpleMixedTest.singletag.tag_model

    def setUpExtra(self):
        self.site = admin.AdminSite(name="tagulous_admin")
        tag_admin.register(self.model, admin_class=self.admin_cls, site=self.site)
        self.ma = self.site._registry[self.model]
        self.cl = None

        # Set up client
        User.objects.create_superuser("test", "test@example.com", "user")
        result = self.client.login(username="test", password="user")
        self.assertEqual(result, True)

    def tearDownExtra(self):
        test_urls.urlpatterns = self.old_urls
        self.client.logout()

    def get_url(self, name, *args, **kwargs):
        content_type = ContentType.objects.get_for_model(self.model)
        return reverse(
            "admin:%s_%s_%s" % (content_type.app_label, content_type.model, name),
            args=args,
            kwargs=kwargs,
        )

    #
    # Tests
    #

    def test_add_renders(self):
        "Check inline add renders without error"
        response = self.client.get(self.get_url("add"))
        self.assertContains(response, "<h2>Simple mixed tests</h2>")
        # assertHTMLEquals isn't going to be particularly helpful here
        # Just check that a few attributes exist, to indicate all is well
        # Real test will be done in next test
        self.assertContains(response, 'id="id_simplemixedtest_set-TOTAL_FORMS')
        self.assertContains(response, 'id="id_simplemixedtest_set-0-singletag"')
        self.assertContains(response, 'id="id_simplemixedtest_set-0-name"')
        self.assertContains(response, 'id="id_simplemixedtest_set-0-tags"')
        # There should be 3 empty fields, zero-indexed
        self.assertContains(response, 'id="id_simplemixedtest_set-2-singletag"')
        self.assertNotContains(response, 'id="id_simplemixedtest_set-3-singletag"')

    def test_add_submits(self):
        "Check inline add saves without error, and increments tag model"
        data = {
            "simplemixedtest_set-TOTAL_FORMS": 3,
            "simplemixedtest_set-INITIAL_FORMS": 0,
            "simplemixedtest_set-MAX_NUM_FORMS": 1000,
            "_save": "Save",
            "name": "Mr",
            "slug": "mr",
            "simplemixedtest_set-0-name": "Test 1",
            "simplemixedtest_set-0-tags": "tag1, tag2",
        }
        response = self.client.post(self.get_url("add"), data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(self.tagged_model.objects.count(), 1)

        t1 = self.tagged_model.objects.get(name="Test 1")
        self.assertInstanceEqual(
            t1, name="Test 1", singletag="Mr", tags=["tag1", "tag2"]
        )
        self.assertTagModel(self.model, {"Mr": 1})
        self.assertTagModel(self.tagged_model.tags.tag_model, {"tag1": 1, "tag2": 1})

    def test_edit_renders(self):
        "Check inline edit renders without error"
        obj1 = self.tagged_model.objects.create(
            name="Test 1", singletag="Mr", tags=["tag1", "tag2"]
        )

        response = self.client.get(self.get_url("change", obj1.singletag.pk))

        self.assertContains(response, "<h2>Simple mixed tests</h2>")
        self.assertContains(response, 'id="id_simplemixedtest_set-TOTAL_FORMS')
        self.assertContains(response, 'id="id_simplemixedtest_set-0-singletag"')
        self.assertContains(response, 'id="id_simplemixedtest_set-0-name"')
        self.assertContains(response, 'id="id_simplemixedtest_set-0-tags"')
        # There should be 1 full field and 3 empty fields, zero-indexed
        self.assertContains(response, 'id="id_simplemixedtest_set-3-singletag"')
        self.assertNotContains(response, 'id="id_simplemixedtest_set-4-singletag"')

        # Check the fk is an id, to confirm it's using TaggedInlineFormSet
        html = response.content.decode("utf-8")
        i_name = html.index('name="simplemixedtest_set-0-singletag"')
        i_open = html.rindex("<", 0, i_name)
        i_close = html.index(">", i_name)
        self.assertHTMLEqual(
            html[i_open : i_close + 1],
            (
                '<input type="hidden" name="simplemixedtest_set-0-singletag" '
                'value="%d" id="id_simplemixedtest_set-0-singletag" />'
            )
            % obj1.singletag.pk,
        )

    def test_edit_saves(self):
        """
        Check inline edit saves without error

        Also check changing tag fields in inlines
        """
        # Create and confirm
        obj1 = self.tagged_model.objects.create(
            name="Test 1", singletag="Mr", tags=["tag1", "tag2"]
        )
        obj2 = self.tagged_model.objects.create(
            name="Test 2", singletag="Mr", tags=["tag3", "tag4"]
        )
        tag1 = obj1.singletag
        self.assertEqual(self.tagged_model.objects.count(), 2)
        self.assertTagModel(self.model, {"Mr": 2})
        self.assertTagModel(
            self.tagged_model.tags.tag_model,
            {"tag1": 1, "tag2": 1, "tag3": 1, "tag4": 1},
        )

        data = {
            "simplemixedtest_set-TOTAL_FORMS": 5,
            "simplemixedtest_set-INITIAL_FORMS": 2,
            "simplemixedtest_set-MAX_NUM_FORMS": 1000,
            "_save": "Save",
            "name": "Mr",
            "slug": "mr",
            "simplemixedtest_set-0-name": "Test 1e",
            "simplemixedtest_set-0-tags": "tag1, tag2e",
            "simplemixedtest_set-0-id": "%d" % obj1.pk,
            "simplemixedtest_set-0-singletag": "%d" % tag1.pk,
            "simplemixedtest_set-0-DELETE": "",
            "simplemixedtest_set-1-name": "Test 2e",
            "simplemixedtest_set-1-tags": "tag3, tag4e",
            "simplemixedtest_set-1-id": "%d" % obj2.pk,
            "simplemixedtest_set-1-singletag": "%d" % tag1.pk,
            "simplemixedtest_set-1-DELETE": "",
        }
        response = self.client.post(self.get_url("change", tag1.pk), data)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(self.tagged_model.objects.count(), 2)

        self.assertInstanceEqual(
            obj1, name="Test 1e", singletag="Mr", tags=["tag1", "tag2e"]
        )
        self.assertInstanceEqual(
            obj2, name="Test 2e", singletag="Mr", tags=["tag3", "tag4e"]
        )

        self.assertTagModel(self.model, {"Mr": 2})
        self.assertTagModel(
            self.tagged_model.tags.tag_model,
            {"tag1": 1, "tag2e": 1, "tag3": 1, "tag4e": 1},
        )



"""
Tagulous tag models
"""
from django.db import IntegrityError, models, router, transaction
from django.db.models import F, Max
from django.db.models.functions import Floor
from django.utils.text import slugify

from .. import constants, settings, utils
from .options import TagOptions


# ##############################################################################
# ###### TagModel manager and queryset
# ##############################################################################


class TagModelQuerySet(models.query.QuerySet):
    def initial(self):
        """
        Reduce the queryset to only include initial tags left in the queryset
        """
        return self.filter(name__in=self.model.tag_options.initial)

    def filter_or_initial(self, *args, **kwargs):
        """
        Reduce the queryset to match the specified filter, but also include
        any initial tags left in the queryset.

        An example of usage would be where tags are only visible to the user
        who added them, but you also want them to see the initial default tags.
        """
        return self.filter(
            models.Q(*args, **kwargs)
            | models.Q(name__in=self.model.tag_options.initial)
        )

    def weight(self, min=settings.WEIGHT_MIN, max=settings.WEIGHT_MAX):
        """
        Add a ``weight`` integer field to objects, weighting the ``count``
        between ``min`` and ``max``.

        Suitable for use with a tag cloud
        """
        # Ignoring PEP 8 intentionally regarding conflict of min/max keywords -
        # concerns are outweighed by clarity of function argument names.

        # Weight is the count scaled to the min/max bounds
        # weight = ( (count * (max - min)) / max_count ) + min
        scale = int(max) - int(min)
        max_count = self.model.objects.aggregate(Max("count"))["count__max"] or 1
        qs = self.annotate(weight=(Floor(F("count") * scale) / max_count) + int(min))
        return qs

    def __str__(self):
        return utils.render_tags(self)


class TagModelManager(models.Manager):
    def get_queryset(self):
        return TagModelQuerySet(self.model, using=self._db)

    get_query_set = get_queryset

    def __str__(self):
        return str(self.get_queryset())

    def initial(self):
        return self.get_queryset().initial()

    def filter_or_initial(self, *args, **kwargs):
        return self.get_queryset().filter_or_initial(*args, **kwargs)

    def weight(self, *args, **kwargs):
        return self.get_queryset().weight(*args, **kwargs)


# ##############################################################################
# ###### Abstract base class for all TagModel models
# ##############################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Metaclass for tag models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagModelBase(models.base.ModelBase):
    def __new__(cls, name, bases, attrs):
        # Set up as normal
        new_cls = super(TagModelBase, cls).__new__(cls, name, bases, attrs)

        # TagMeta takes priority for the model
        new_tag_options = None
        if "TagMeta" in attrs:
            tag_meta = {}
            tag_meta = dict(
                (key, val)
                for key, val in attrs["TagMeta"].__dict__.items()
                if key in constants.OPTION_DEFAULTS
            )
            if "tree" in tag_meta:
                raise ValueError("Cannot set tree option in TagMeta")

            new_tag_options = TagOptions(**tag_meta)

        # Failing that, look for a direct tag_options setting
        # It will probably have been passed by BaseTagField.contribute_to_class
        elif "tag_options" in attrs:
            new_tag_options = attrs["tag_options"]

        # Otherwise start a new one
        else:
            new_tag_options = TagOptions()

        # See if there's anything to inherit
        # This also means that tag_options will be available on abstract models
        if hasattr(new_cls, "tag_options"):
            # Inherit by setting missing values in place
            new_tag_options.set_missing(new_cls.tag_options)

        # Assign
        new_cls.tag_options = new_tag_options

        # Check for self-referential tag fields on this model
        fields = new_cls._meta.fields + new_cls._meta.many_to_many

        for field in fields:
            # Can't test for subclass of field here - would be circular import
            if hasattr(field, "tag_model") and field.tag_model == new_cls:
                # This method is being called after the tag field's
                # contribute_to_class and _process_deferred_options. This means
                # that the field is using tag_options from the inherited model.
                # Change it to use this one
                field.tag_options = new_tag_options

        return new_cls


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Empty abstract model
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class BaseTagModel(models.Model, metaclass=TagModelBase):
    """
    Empty abstract base class for tag models

    This is used when dynamically building models, eg in migrations
    """

    objects = TagModelManager()

    class Meta:
        abstract = True

    def __str__(self):
        return str(self.name)

    def __eq__(self, obj):
        """
        If comparing to a string, is equal if string value matches
        Otherwise compares normally
        """
        if isinstance(obj, str):
            return self.name == obj
        return super(BaseTagModel, self).__eq__(obj)

    __hash__ = models.Model.__hash__

    def __ne__(self, obj):
        return not self == obj

    def get_absolute_url(self):
        if self.tag_options.get_absolute_url is None:
            raise AttributeError(
                "'%s' has no attribute 'get_absolute_url'" % self.__class__.__name__
            )
        return self.tag_options.get_absolute_url(self)

    @classmethod
    def get_related_fields(cls, include_standard=False):
        """
        Return a list of related tag fields in other models which refer to this
        tag model.

        If include_standard=False (default), only SingleTagFields and
        TagFields will be returned. If True, it will also include ForeignKeys
        and ManyToManyFields.
        """
        # Avoid circular import
        from .fields import SingleTagField, TagField

        meta = cls._meta

        related_fields = [
            f
            for f in meta.get_fields()
            if (f.many_to_many or f.one_to_many or f.one_to_one) and f.auto_created
        ]

        if include_standard:
            return related_fields

        return [
            f
            for f in related_fields
            if isinstance(f.field, (SingleTagField, TagField,),)
        ]

    def get_related_objects(self, flat=False, distinct=False, include_standard=False):
        """
        Get any instances of other models which refer to this tag instance

        If flat=False, returns the following data structure:
            [
                [related_model, related_field, [instance, instance, ...],
                [related_model, related_field, [instance, instance, ...],
                ...
            ]

        If flat=False, the distinct argument is ignored. Models and fields
        will only be included if they refer to this specific tag - there will
        not be any empty lists of instances.

        If flat=True, returns a list of instances without any information about
        how they are related:
            [ instance, instance, ... ]

        If flat=True and distinct=True, only unique instances will be returned.

        If include_standard=False (default), only SingleTagFields and
        TagFields will be returned. If True, it will also include ForeignKeys
        and ManyToManyFields.
        """
        data = []
        for related in self.get_related_fields(include_standard=include_standard):
            related_model = related.related_model

            objs = related_model._base_manager.using(
                router.db_for_write(self.tag_model)
            ).filter(**{"%s" % related.field.name: self})
            if not objs:
                continue
            if flat:
                data.extend(objs)
            else:
                data.append([related_model, related.field, objs])
        if flat and distinct:
            data = list(set(data))
        return data

    def update_count(self):
        """
        Count how many SingleTagFields and TagFields refer to this tag, save,
        and try to delete.
        """
        self.count = len(self.get_related_objects(flat=True))
        self.save()
        self.try_delete()

    update_count.alters_data = True

    def increment(self):
        """
        Increase the count by one
        """
        self._change_count(1)

    increment.alters_data = True

    def decrement(self):
        """
        Decrease the count by one, then try to delete
        """
        self._change_count(-1)

    decrement.alters_data = True

    def _change_count(self, amount):
        """
        Change count by amount
        """
        self.__class__.objects.filter(pk=self.pk).update(
            count=models.F("count") + amount
        )

        # Reload count
        # Use DB for write because we just updated the value
        using = router.db_for_write(self.tag_model, instance=self)
        self.refresh_from_db(using=using)

        self.try_delete()

    def try_delete(self):
        """
        If count is 0, try to delete this tag
        """
        if self.count != 0:
            return

        # See if it's protected
        is_protected = self.protected or self.tag_options.protect_all
        if not is_protected:
            # Before we delete, check for any standard relationships
            # This will catch if the tag is in a tree with children
            true_count = len(self.get_related_objects(flat=True, include_standard=True))
            if true_count > 0:
                # ForeignKeys or ManyToManyFields refer to it
                # We can't delete (we'll break things)
                # Tag is protected, for now
                is_protected = True

        # Try to delete
        if not is_protected:
            # Tag is not in use and not protected. Delete.
            self.delete()

            # If a tree, parent node may now be empty - try to delete it
            if self.tag_options.tree and self.parent:
                self.parent.try_delete()

    try_delete.alters_data = True

    def _prep_merge_tags(self, tags):
        """
        Ensure tags argument for merge_tags is an iterable of tag objects,
        excluding self
        """
        # Ensure tags is a list of tag instances
        if isinstance(tags, str):
            tags = utils.parse_tags(
                tags, space_delimiter=self.tag_options.space_delimiter
            )
        if not isinstance(tags, models.query.QuerySet):
            tags = self.tag_model.objects.filter(name__in=tags)

        # Make sure self isn't in tags
        return tags.exclude(pk=self.pk)

    def merge_tags(self, tags):
        """
        Merge the specified tags into this tag
        """
        # Avoid circular import
        from .fields import SingleTagField, TagField

        related_fields = self.tag_model.get_related_fields()
        tags = self._prep_merge_tags(tags)

        for related in related_fields:
            # Get the instances of the related models which refer to the tag
            # instances being merged
            related_model = related.related_model

            objs = related_model._base_manager.using(
                router.db_for_write(self.tag_model, instance=self)
            ).filter(**{"%s__in" % related.field.name: tags})

            # Switch the tags
            if isinstance(related.field, SingleTagField):
                for obj in objs:
                    setattr(obj, related.field.name, self)
                    obj.save()

            elif isinstance(related.field, TagField):
                for obj in objs:
                    getattr(obj, related.field.name).remove(*tags)
                    getattr(obj, related.field.name).add(self)

    merge_tags.alters_data = True

    def _update_extra(self):
        """
        Called by .save() before super().save()

        Allows subclasses to update extra fields based on slug
        """
        pass

    def _save_direct(self, *args, **kwargs):
        """
        Save without modifying data

        Used during schemamigrations
        """
        # We inherited from BaseTagModel, tell that to save directly too
        return super(BaseTagModel, self).save(*args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Automatically generate a unique slug, if one does not exist
        """
        # Based on django-taggit: don't worry about race conditions when
        # setting names and slugs, just avoid potential slugify clashes.
        # We could improve this if race conditions are ever a problem in the
        # real world, but until Django provides a reliable way to determine
        # the cause of an IntegrityError, we can never make this perfect.

        # If already in the database and has a slug set, just save as normal
        # Set slug to None to rebuild it
        if self.pk and self.slug:
            self._update_extra()
            return super(BaseTagModel, self).save(*args, **kwargs)

        # Set the slug using the label if possible (for TagTreeModel), else
        # the tag name.
        label = getattr(self, "label", self.name)
        slug_max_length = self.__class__._meta.get_field("slug").max_length
        if settings.SLUG_ALLOW_UNICODE:
            slug_base = slugify(label, allow_unicode=True)
        else:
            slug_base = slugify(label, allow_unicode=False)

            # Django 3.2 strips trailing and leading underscores; this risks creating an
            # empty slug for unconvertable characters, eg logographic characters. Ensure
            # they are not empty.
            if slug_base == "":
                slug_base = "_"

        # ASCII-ification can make a longer string
        self.slug = slug_base[:slug_max_length]
        self._update_extra()

        # Make sure we're using the same db at all times
        cls = self.__class__
        kwargs["using"] = kwargs.get("using") or router.db_for_write(cls, instance=self)

        # Try saving the slug - it'll probably be fine
        try:
            # If transaction supports atomic, we need to wrap the save call -
            # otherwise if save throws an exception it'll cause any current
            # queries to roll back
            with transaction.atomic():
                return super(BaseTagModel, self).save(*args, **kwargs)
        except IntegrityError:
            pass

        # Integrity error - something is probably not unique.
        # Assume it was the slug - make it unique by appending a number.
        # See which numbers have been used
        slug_base = slug_base[: slug_max_length - settings.SLUG_TRUNCATE_UNIQUE]
        try:
            last = cls.objects.filter(slug__regex="^%s_[0-9]+$" % slug_base).latest(
                "slug"
            )
        except cls.DoesNotExist:
            # No numbered version of the slug exists
            number = 1
        else:
            slug_base, number = last.slug.rsplit("_", 1)
            number = int(number) + 1

        self.slug = "%s_%d" % (slug_base, number)
        self._update_extra()
        return super(BaseTagModel, self).save(*args, **kwargs)

    save.alters_data = True

    # For consistency with SingleTagField, provide .tag_model attribute
    tag_model = property(lambda self: self.__class__)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Abstract model with fields
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagModel(BaseTagModel):
    """
    Abstract base class for tag models
    """

    name = models.CharField(unique=True, max_length=settings.NAME_MAX_LENGTH)
    slug = models.SlugField(unique=False, max_length=settings.SLUG_MAX_LENGTH)
    count = models.IntegerField(
        default=0, help_text="Internal counter of how many times this tag is in use"
    )
    protected = models.BooleanField(
        default=False, help_text="Will not be deleted when the count reaches 0"
    )

    class Meta:
        abstract = True
        ordering = ("name",)
        # Slug field must be unique, but manage it with unique_together
        # so that subclasses can override
        unique_together = (("slug",),)


# ##############################################################################
# ###### TagTreeModel manager and queryset
# ##############################################################################


class TagTreeModelQuerySet(TagModelQuerySet):
    def _clean(self):
        """
        Return a clean copy of this class
        """
        kwargs = {"model": self.model, "using": self._db}
        if hasattr(self, "_hints"):
            kwargs["hints"] = self._hints
        return self.__class__(**kwargs)

    def with_ancestors(self):
        """
        Add selected tags' ancestors to current queryset
        """
        # Build list of all paths of all ancestors (and self)
        paths = []
        for path in self.values_list("path", flat=True):
            parts = utils.split_tree_name(path)
            paths += [path] + [
                utils.join_tree_name(parts[:i])  # Join parts up to i (misses last)
                for i in range(1, len(parts))  # Skip first (empty)
            ]
        return self._clean().filter(path__in=set(paths))

    def with_descendants(self):
        """
        Add selected tags' descendants to current queryset
        """
        # Build query of all matching paths, and all their sub-paths
        query = models.Q()
        for path in self.values_list("path", flat=True):
            query = (
                query | models.Q(path=path) | models.Q(path__startswith="%s/" % path)
            )
        return self._clean().filter(query)

    def with_siblings(self):
        """
        Add selected tags' siblings to current queryset
        """
        # Get all unique parent pks, except None
        parent_ids = set(self.values_list("parent_id", flat=True))
        has_none = None in parent_ids
        if has_none:
            parent_ids.remove(None)

        # If None is there, we need to test with isnull
        query = models.Q(parent_id__in=list(parent_ids))
        if has_none:
            query = query | models.Q(parent_id__isnull=True)

        return self._clean().filter(query)


class TagTreeModelManager(TagModelManager):
    def get_queryset(self):
        return TagTreeModelQuerySet(self.model, using=self._db)

    get_query_set = get_queryset

    def rebuild(self):
        """
        Re-save each instance to update tag fields
        """
        for tag in self.all():
            # Replace slug in case name has changed
            # If it hasn't, it'll just end up re-creating the same one
            tag.slug = None
            tag.save()

    rebuild.alters_data = True


# ##############################################################################
# ###### Abstract base class for all TagTreeModel models
# ##############################################################################

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Metaclass for tag tree models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagTreeModelBase(TagModelBase):
    def __new__(cls, name, bases, attrs):
        # Set up as normal
        new_cls = super(TagTreeModelBase, cls).__new__(cls, name, bases, attrs)

        # Force tree tag_options to True
        new_cls.tag_options.tree = True

        return new_cls


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Empty abstract model
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class BaseTagTreeModel(BaseTagModel, metaclass=TagTreeModelBase):
    """
    Empty abstract base class for tag models with tree

    This is used when dynamically building models, eg in migrations
    """

    objects = TagTreeModelManager()

    class Meta:
        abstract = True

    # Other derivable attributes won't be used in lookups, so don't need to be
    # cached. If there are situations where they are needed for lookups, this
    # model can be subclassed (or better yet, use a reusable mixin) and the
    # properties replaced by caching fields.
    def _get_descendant_count(self):
        "The sum of the counts of all descendants"
        return self.get_descendants().aggregate(models.Sum("count"))["count__sum"] or 0

    descendant_count = property(
        _get_descendant_count, doc=_get_descendant_count.__doc__
    )

    def _get_family_count(self):
        "The count of self plus all its descendants"
        return self.descendant_count + self.count

    family_count = property(_get_family_count, doc=_get_family_count.__doc__)

    def __init__(self, *args, **kwargs):
        """
        Initialise the tag
        """
        super(BaseTagTreeModel, self).__init__(*args, **kwargs)
        # Keep track of the name
        self._name = self.name

    def _save_direct(self, *args, **kwargs):
        """
        Save without modifying data

        Used during schemamigrations
        """
        # We inherited from BaseTagModel, tell that to save directly too
        return super(BaseTagTreeModel, self)._save_direct(*args, **kwargs)

    def save(self, *args, **kwargs):
        """
        Set the parent and path cache
        """
        # Make sure name is valid
        if self.name is None or not self.name.strip():
            raise IntegrityError(
                "NOT NULL constraint failed: {}.name".format(self._meta.db_table)
            )
        self.name = utils.clean_tree_name(self.name)

        # Find the parent, or create it if missing
        parts = utils.split_tree_name(self.name)
        old_parent = self.parent
        if len(parts) > 1:
            self.parent, created = self.__class__.objects.get_or_create(
                name=utils.join_tree_name(parts[:-1])
            )
        else:
            self.parent = None

        # Update other cache fields
        self.label = parts[-1]
        self.level = len(parts)

        # Save - super .save() method will set the path using _get_path()
        super(BaseTagTreeModel, self).save(*args, **kwargs)

        # If name has changed...
        if self._name != self.name:
            # Update child names
            for child in self.children.all():
                child.name = utils.join_tree_name(parts + [child.label])
                child.save()
            self._name = self.name

            # Notify parent that it may now be empty
            if old_parent:
                old_parent.update_count()

    save.alters_data = True

    def _update_extra(self):
        """
        Updates extra fields based on slug
        """
        # Update the path
        if self.parent:
            self.path = "/".join([self.parent.path, self.slug])
        else:
            self.path = self.slug

    _update_extra.alters_data = True

    def merge_tags(self, tags, children=False):
        """
        Merge the specified tags into this tag

        If children is True, child tags will also be merged into children of
        this tag (retaining structure)
        """
        tags = self._prep_merge_tags(tags)

        # Merge children first
        if children:
            cls = self.__class__
            for tag in tags:
                for child in tag.children.all():
                    # Find merge target on this tree
                    new_child_name = "/".join([self.name, child.label])
                    try:
                        my_child = cls.objects.get(name=new_child_name)
                    except cls.DoesNotExist:
                        my_child = None

                    if my_child:
                        # It exists - merge recursively
                        my_child.merge_tags([child], children=True)
                    else:
                        # It doesn't exist - rename recursively
                        child.name = new_child_name
                        child.save()

        # Merge self
        super(BaseTagTreeModel, self).merge_tags(tags)

    merge_tags.alters_data = True

    def get_ancestors(self):
        """
        Get a queryset of ancestors for this tree node
        """
        cls = self.__class__
        if not self.parent:
            return cls.objects.none()

        # Get all ancestor paths from this path
        parts = utils.split_tree_name(self.path)
        paths = [
            utils.join_tree_name(parts[:i])  # Join parts up to i (misses last)
            for i in range(1, len(parts))  # Skip first (empty)
        ]

        # Look up ancestors by path, already ordered by name for deepest last
        return cls.objects.filter(path__in=paths)

    def get_descendants(self):
        """
        Get a queryset of descendants of this tree node
        """
        # Look up by path, already ordered by name for deepest last
        cls = self.__class__
        return cls.objects.filter(path__startswith="%s/" % self.path)

    def get_siblings(self):
        """
        Get a queryset of siblings of this tree node, including this node.

        If you don't want this node in the results, exclude it afterwards, eg:

            node.get_siblings().exclude(pk=node.pk)
        """
        if self.parent:
            return self.parent.children.all()
        return self.__class__.objects.filter(level=1)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#       Abstract model with fields
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


class TagTreeModel(BaseTagTreeModel, TagModel):
    """
    Abstract base class for tag models with tree
    """

    # These fields are all generated automatically on save
    parent = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        related_name="children",
        on_delete=models.CASCADE,
        db_index=True,
    )
    path = models.TextField()
    label = models.CharField(
        max_length=settings.LABEL_MAX_LENGTH,
        help_text="The name of the tag, without ancestors",
    )
    level = models.IntegerField(default=1, help_text="The level of the tag in the tree")

    class Meta:
        abstract = True
        ordering = ("name",)
        unique_together = (("slug", "parent"),)



"""
Python serializer with Tagulous support
"""
from django.core.serializers import python as python_serializer

from . import base


class Serializer(base.SerializerMixin, python_serializer.Serializer):
    """
    Serializes a QuerySet to basic Python objects, with tag field support
    """

    pass


Deserializer = base.DeserializerWrapper(
    python_serializer.Deserializer,
    doc=(
        "Deserialize Python objects into Django ORM instances, with tag field "
        "support"
    ),
)

base.monkeypatch_get_model(python_serializer)



from django.apps import AppConfig


class TagulousConfig(AppConfig):
    name = "tagulous"

    def ready(self):
        from .signals.post import register_post_signals

        register_post_signals()
